""" Import data to PostgreSQL database """

import getpass

import sqlalchemy
import sqlalchemy.engine.reflection
import sqlalchemy.engine.url
import sqlalchemy_utils
from pyhelpers.ops import confirmed


class RailwayCodesPSQL:
    def __init__(self, username='postgres', password=None, host='localhost', port=5432, database_name='postgres'):
        """
        We need to be connected to the database server in order to execute the "CREATE DATABASE" command. There is a 
        database called "postgres" created by the "initdb" command when the data storage area is initialised. If we 
        need to create the first of our own database, we can set up a connection to "postgres" in the first instance.
        """
        self.database_info = {'drivername': 'postgresql+psycopg2',
                              'username': username,
                              'password': password if password else getpass.getpass('PostgreSQL password: '),
                              'host': host,  # default: localhost
                              'port': port,  # 5432 (default by installation)
                              'database': database_name}

        # The typical form of a database URL is: url = backend+driver://username:password@host:port/database_name
        self.url = sqlalchemy.engine.url.URL(**self.database_info)
        self.dialect = self.url.get_dialect()
        self.backend = self.url.get_backend_name()
        self.driver = self.url.get_driver_name()
        self.user, self.host = self.url.username, self.url.host
        self.port = self.url.port
        self.database_name = self.database_info['database']

        # Create a SQLAlchemy connectable
        self.engine = sqlalchemy.create_engine(self.url, isolation_level='AUTOCOMMIT')
        self.connection = self.engine.connect()

    # Establish a connection to the specified database (named e.g. 'osm_extracts')
    def connect_db(self, database_name='Railway_Codes'):
        """
        :param database_name: [str] (default: 'Railway_Codes') name of a database
        """
        self.database_name = database_name
        self.database_info['database'] = self.database_name
        self.url = sqlalchemy.engine.url.URL(**self.database_info)
        if not sqlalchemy_utils.database_exists(self.url):
            sqlalchemy_utils.create_database(self.url)
        self.engine = sqlalchemy.create_engine(self.url, isolation_level='AUTOCOMMIT')
        self.connection = self.engine.connect()

    # Check if a database exists
    def db_exists(self, database_name):
        result = self.engine.execute("SELECT EXISTS("
                                     "SELECT datname FROM pg_catalog.pg_database "
                                     "WHERE datname='{}');".format(database_name))
        return result.fetchone()[0]

    # An alternative to sqlalchemy_utils.create_database()
    def create_db(self, database_name='Railway Codes and other data'):
        """
        :param database_name: [str, 'Railway Codes and other data'(default)] name of a database

        from psycopg2 import OperationalError
        try:
            self.engine.execute('CREATE DATABASE "{}"'.format(database_name))
        except OperationalError:
            self.engine.execute(
                'SELECT *, pg_terminate_backend(pid) FROM pg_stat_activity WHERE username=\'postgres\';')
            self.engine.execute('CREATE DATABASE "{}"'.format(database_name))
        """
        if not self.db_exists(database_name):
            self.disconnect()
            self.engine.execute('CREATE DATABASE "{}";'.format(database_name))
            self.connect_db(database_name=database_name)
        else:
            print("The database already exists.")

    # Get size of a database
    def get_db_size(self, database_name=None):
        """
        :param database_name: [str; None (default)] name of database
        :return:
        """
        db_name = '\'{}\''.format(database_name) if database_name else 'current_database()'
        db_size = self.engine.execute('SELECT pg_size_pretty(pg_database_size({})) AS size;'.format(db_name))
        print(db_size.fetchone()[0])

    # Kill the connection to the specified database
    def disconnect(self, database_name=None):
        """
        :param database_name: [str; None(default)] name of database to disconnect from

        Alternative way:
        SELECT
            pg_terminate_backend(pg_stat_activity.pid)
        FROM
            pg_stat_activity
        WHERE
            pg_stat_activity.datname = database_name AND pid <> pg_backend_pid();
        """
        db_name = self.database_name if database_name is None else database_name
        self.connect_db('postgres')
        self.engine.execute('REVOKE CONNECT ON DATABASE {} FROM PUBLIC, postgres;'.format(db_name))
        self.engine.execute(
            'SELECT pg_terminate_backend(pid) '
            'FROM pg_stat_activity '
            'WHERE datname = \'{}\' AND pid <> pg_backend_pid();'.format(db_name))

    # Kill connections to all other databases
    def disconnect_all_others(self):
        self.connect_db('postgres')
        self.engine.execute('SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE pid <> pg_backend_pid();')

    # Drop the specified database
    def drop(self, database_name=None):
        """
        :param database_name: [str; None (default)] database to be disconnected; None: to disconnect the current one
        """
        db_name = self.database_name if database_name is None else database_name
        if confirmed("Confirmed to drop the database \"{}\"?".format(db_name)):
            self.disconnect(db_name)
            self.engine.execute('DROP DATABASE IF EXISTS "{}"'.format(db_name))

    # Create a new schema in the database being currently connected
    def create_schema(self, schema_name):
        """
        :param schema_name: [str] name of a schema
        """
        self.engine.execute('CREATE SCHEMA IF NOT EXISTS "{}";'.format(schema_name))

    # Drop a schema in the database being currently connected
    def drop_schema(self, *schema_names):
        """
        :param schema_names: [str] name of one schema, or names of multiple schemas
        """
        if schema_names:
            schemas = tuple(schema_name for schema_name in schema_names)
        else:
            schemas = tuple(
                x for x in sqlalchemy.engine.reflection.Inspector.from_engine(self.engine).get_schema_names()
                if x != 'public' and x != 'information_schema')
        if confirmed("Confirmed to drop the schema(s): {}".format(schemas)):
            self.engine.execute(('DROP SCHEMA IF EXISTS ' + '%s, '*(len(schemas) - 1) + '%s CASCADE;') % schemas)

    # Check if a table exists
    def table_exists(self, schema_name, table_name):
        """
        :param schema_name: [str] name of a schema
        :param table_name: [str] name of a table
        :return: [bool]
        """
        result = self.engine.execute("SELECT EXISTS("
                                     "SELECT * FROM information_schema.tables "
                                     "WHERE table_schema='{}' "
                                     "AND table_name='{}');".format(schema_name, table_name))
        return result.fetchone()[0]

    # Import data (as a pandas.DataFrame) into the database being currently connected
    def dump_data(self, data, schema_name, table_name, if_exists='replace', chunk_size=None):
        """
        :param data: [pandas.DataFrame]
        :param schema_name: [str]
        :param table_name: [str] name of the targeted table
        :param if_exists: [str] 'fail', 'replace', or 'append'; default 'replace'
        :param chunk_size: [int; None]
        """
        if schema_name not in sqlalchemy.engine.reflection.Inspector.from_engine(self.engine).get_schema_names():
            self.create_schema(schema_name)

        if not data.empty:
            data.coordinates = data.coordinates.map(lambda x: x.wkt)
            data.other_tags = data.other_tags.astype(str)
        data.to_sql(table_name, self.engine, schema=schema_name, if_exists=if_exists, index=False, chunksize=chunk_size)

    # Remove tables from the database being currently connected
    def drop_data(self, schema_name, *table_names):
        """
        :param schema_name: [str]
        :param table_names: [str]
        """
        tables = tuple(('{}.\"{}\"'.format(schema_name, table_name) for table_name in table_names))
        self.engine.execute(('DROP TABLE IF EXISTS ' + '%s, '*(len(tables) - 1) + '%s CASCADE;') % tables)



"""
Data source: http://www.railwaycodes.org.uk

Engineer's Line References (ELRs) (http://www.railwaycodes.org.uk/elrs/elr0.shtm)

"Mileages are given in the form miles.chains. Figures prefixed with a tilde (~) are approximate, items in parentheses
are not on this route but are given for reference."
"""

import copy
import itertools
import os
import re
import string

import bs4
import measurement.measures
import numpy as np
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.ops import confirmed
from pyhelpers.store import load_pickle

from pyrcs.utils import cd_dat, get_catalogue, get_last_updated_date, parse_table
from pyrcs.utils import mile_chain_to_nr_mileage, nr_mileage_to_mile_chain, yards_to_nr_mileage
from pyrcs.utils import save_pickle

# ====================================================================================================================
""" Utilities """


# Check if a str expresses a float
def is_str_float(str_val):
    try:
        float(str_val)  # float(re.sub('[()~]', '', text))
        test_res = True
    except ValueError:
        test_res = False
    return test_res


#
def identify_multiple_measures(mileage_data):
    """
    e.g. elr='MLA', elr='FED'
    """
    test_temp = mileage_data[~mileage_data.Mileage.astype(bool)]
    if not test_temp.empty:
        test_temp_node, sep_rows_idx = test_temp.Node.tolist(), test_temp.index[-1]
        if '1949 measure' in test_temp_node:
            mileage_data.Node = mileage_data.Node.str.replace('1949 measure', 'Current measure')
            test_temp_node = [re.sub(r'1949 ', 'Current ', x) for x in test_temp_node]
        if 'Distances in km' in test_temp_node:
            temp_mileage_data = mileage_data[~mileage_data.Node.str.contains('Distances in km')]
            temp_mileages = temp_mileage_data.Mileage.map(
                lambda x: nr_mileage_to_mile_chain(yards_to_nr_mileage(measurement.measures.Distance(km=x).yd)))
            temp_mileage_data.Mileage = temp_mileages.tolist()
            checked_mileage_data = temp_mileage_data
        elif 'One measure' in test_temp_node:
            sep_rows_idx = mileage_data[mileage_data.Node.str.contains('Alternative measure')].index[0]
            mileage_data_1, mileage_data_2 = np.split(mileage_data, [sep_rows_idx], axis=0)
            checked_mileage_data = {
                'One measure': mileage_data_1[~mileage_data_1.Node.str.contains('One measure')],
                'Alternative measure': mileage_data_2[~mileage_data_2.Node.str.contains('Alternative measure')]}
        elif 'This line has two \'legs\':' in test_temp_node:
            temp_mileage_data = mileage_data.iloc[1:].drop_duplicates()
            temp_mileage_data.index = range(len(temp_mileage_data))
            checked_mileage_data = temp_mileage_data
        else:
            test_temp_text = [' '.join(x) for x in itertools.product(
                *(('Current', 'Later', 'One', 'Original', 'Former', 'Alternative', 'Usual', 'Earlier'),
                  ('measure', 'route')))]
            alt_sep_rows_idx = [x in test_temp_node for x in test_temp_text]
            num_of_measures = sum(alt_sep_rows_idx)
            if num_of_measures == 1:  #
                mileage_data_1, mileage_data_2 = np.split(mileage_data, [sep_rows_idx], axis=0)
                if re.match(r'(Original)|(Former)|(Alternative)|(Usual)', test_temp_node[0]):
                    measure_ = re.sub(r'(Original)|(Former)|(Alternative)|(Usual)', r'Current',
                                      test_temp_node[0])
                else:
                    measure_ = re.sub(r'(Current)|(Later)|(One)', r'Previous', test_temp_node[0])
                checked_mileage_data = {measure_: mileage_data_1.loc[0:sep_rows_idx, :],
                                        test_temp_node[0]: mileage_data_2.loc[sep_rows_idx + 1:, :]}
            elif num_of_measures == 2:  # e.g. elr='BTJ'
                sep_rows_idx_items = [test_temp_text[x] for x in np.where(alt_sep_rows_idx)[0]]
                sep_rows_idx = mileage_data[mileage_data.Node.isin(sep_rows_idx_items)].index[-1]
                mileage_data_1, mileage_data_2 = np.split(mileage_data, [sep_rows_idx], axis=0)
                sep_rows_idx_items_checked = [
                    mileage_data_1[mileage_data_1.Node.isin(sep_rows_idx_items)].Node.iloc[0],
                    mileage_data_2[mileage_data_2.Node.isin(sep_rows_idx_items)].Node.iloc[0]]
                mileage_data_1 = mileage_data_1[~mileage_data_1.Node.isin(sep_rows_idx_items)]
                mileage_data_2 = mileage_data_2[~mileage_data_2.Node.isin(sep_rows_idx_items)]
                checked_mileage_data = dict(zip(sep_rows_idx_items_checked, [mileage_data_1, mileage_data_2]))
            else:
                if mileage_data.loc[sep_rows_idx, 'Mileage'] == '':
                    mileage_data.loc[sep_rows_idx, 'Mileage'] = mileage_data.loc[sep_rows_idx - 1, 'Mileage']
                checked_mileage_data = mileage_data
    else:
        checked_mileage_data = mileage_data
    return checked_mileage_data


#
def parse_mileage_col(mileage):
    mileage.index = range(len(mileage))
    if any(mileage.str.match('.*km')):
        if all(mileage.str.match('.*km')):
            temp_mileage = mileage.str.replace('km', '').map(
                lambda x: yards_to_nr_mileage(measurement.measures.Distance(km=x.replace('≈', '')).british_yd))
            miles_chains = temp_mileage.map(lambda x: nr_mileage_to_mile_chain(x))  # Might be wrong!
        else:
            miles_chains = mileage.map(lambda x: re.sub(r'/?\d+\.\d+km/?', '', x))
            temp_mileage = miles_chains.map(lambda x: mile_chain_to_nr_mileage(x))
        mileage_note = [x + ' (Approximate)' if x.startswith('≈') else x for x in list(mileage)]
    else:
        if all(mileage.map(is_str_float)):
            temp_mileage = mileage
            mileage_note = [''] * len(temp_mileage)
        else:
            temp_mileage, mileage_note = [], []
            for m in mileage:
                if m == '':
                    temp_mileage.append(m)
                    mileage_note.append('Unknown')
                elif m.startswith('(') and m.endswith(')'):
                    temp_mileage.append(re.search(r'\d+\.\d+', m).group(0))
                    mileage_note.append('Not on this route but given for reference')
                elif m.startswith('≈') or m.endswith('?'):
                    temp_mileage.append(m.strip('≈').strip('?'))
                    mileage_note.append('Approximate')
                elif re.match(r'\d+\.\d+/\s?\d+\.\d+', m):
                    m1, m2 = m.split('/')
                    temp_mileage.append(m1)
                    mileage_note.append(m2.strip() + ' (Alternative)')
                else:
                    temp_mileage.append(m.strip(' ').replace(' ', '.'))
                    mileage_note.append('')
        miles_chains = temp_mileage.copy()
        temp_mileage = [mile_chain_to_nr_mileage(m) for m in temp_mileage]
    parsed_mileage = pd.DataFrame({'Mileage': temp_mileage,
                                   'Mileage_Note': mileage_note,
                                   'Miles_Chains': miles_chains})
    return parsed_mileage


#
def parse_node_col(node):
    #
    def preprocess_node_x(node_x):
        # node_x = node_x.replace(' with Freightliner terminal', ' & Freightliner Terminal'). \
        #     replace(' with curve to', ' with'). \
        #     replace(' (0.37 long)', '')
        # pat = re.compile(r'\w+.*( \(\d+\.\d+\))?(/| and \w+)? with ([A-Z]){3}(\d)?( \(\d+\.\d+\))?')
        pat = re.compile(r'\w+.*( \(\d+\.\d+\))?(/| and \w+)? with ([A-Z]).*(\d)?( \(\d+\.\d+\))?')
        if re.match(pat, node_x):
            node_name = [x.group() for x in re.finditer(r'\w+.*(?= with)', node_x)]
            conn_node = [x.group() for x in re.finditer(r'(?<= with )[^*]+', node_x)]
        else:
            node_name, conn_node = [node_x], [None]
        return node_name + conn_node

    prep_node = pd.DataFrame((preprocess_node_x(n) for n in node), columns=['Node', 'Connection'])

    #
    def parse_nodes(prep_nodes):
        conn_node_lst = []
        for n in prep_nodes.Connection:
            if n is not None:
                if re.match(r'[A-Z]{3}(\d)?( \(\d+.\d+\))? ?/ ?[A-Z]{3}(\d)?( \(\d+.\d+\))?', n):
                    m = [x.strip() for x in n.split('/')]
                else:
                    m = n.split(' and ')
                if len(m) > 2:
                    m = [' and '.join(m[:2]), ' and '.join(m[2:])]
            else:
                m = [n]
            conn_node_lst.append(m)
        #
        assert isinstance(conn_node_lst, list)
        for i in [conn_node_lst.index(c) for c in conn_node_lst if len(c) > 1]:
            temp_lst = [x.replace('later ', '').rstrip(',').split(' and ') for x in conn_node_lst[i]
                        if isinstance(x, str)]
            conn_node_lst[i] = [v for lst in temp_lst for v in lst]
            temp_lst = [x.split(', ') for x in conn_node_lst[i]]
            conn_node_lst[i] = [v for lst in temp_lst for v in lst]
        most_conn = max(len(c) for c in conn_node_lst)
        # conn_node_list = [c + [None] * (most_conn - len(c)) for c in conn_node_list]
        return pd.DataFrame(conn_node_lst, columns=['Link_{}'.format(n + 1) for n in range(most_conn)])

    conn_nodes = parse_nodes(prep_node)

    #
    def uncouple_elr_mileage(node_x):
        # e.g. x = 'ECM5 (44.64)' or x = 'DNT'
        if node_x is None:
            y = ['', '']
        else:
            # pat0 = re.compile(r'\w+.*(( lines)|( terminal))$')
            pat1 = re.compile(r'([A-Z]{3}(\d)?$)|((\w\s?)*\w$)')
            pat2 = re.compile(r'([A-Z]{3}(\d)?$)|(([\w\s&]?)*(\s\(\d+\.\d+\))?$)')
            pat3 = re.compile(r'[A-Z]{3}(\d)?(\s\(\d+.\d+\))?\s\[.*?\]$')
            pat4 = re.compile(r'[A-Z]{3}(\d)?\s\(\d+\.\d+km\)')
            # if re.match(pat0, node_x):
            #     y = ['', '']
            if re.match(pat1, node_x):
                y = [node_x, '']
            elif re.match(pat2, node_x):
                y = [z[:-1] if re.match(r'\d+.\d+\)', z) else z.strip() for z in node_x.split('(')]
                y[0] = '' if len(y[0]) > 4 else y[0]
            elif re.match(pat3, node_x):
                try:
                    y = [re.search(r'[A-Z]{3}(\d)?', node_x).group(0), re.search(r'\d+\.\d+', node_x).group(0)]
                except AttributeError:
                    y = [re.search(r'[A-Z]{3}(\d)?', node_x).group(0), '']
            elif re.match(pat4, node_x):
                y = [re.search(r'[A-Z]{3}(\d)?', node_x).group(0),
                     nr_mileage_to_mile_chain(yards_to_nr_mileage(
                         measurement.measures.Distance(km=re.search(r'\d+\.\d+', node_x).group(0)).yd))]
            else:
                y = [node_x, ''] if len(node_x) <= 4 else ['', '']
            y[0] = y[0] if len(y[0]) <= 4 else ''
        return y

    #
    link_cols = [x for x in conn_nodes.columns if re.match(r'^(Link_\d)', x)]
    link_nodes = conn_nodes[link_cols].applymap(lambda x: uncouple_elr_mileage(x))
    link_elr_mileage = pd.concat(
        [pd.DataFrame(link_nodes[col].values.tolist(), columns=[col + '_ELR', col + '_Mile_Chain'])
         for col in link_cols], axis=1, sort=False)
    parsed_node_and_conn = pd.concat([prep_node, conn_nodes, link_elr_mileage], axis=1, sort=False)

    return parsed_node_and_conn


#
def parse_mileage_data(mileage_data):
    mileage, node = mileage_data.iloc[:, 0], mileage_data.iloc[:, 1]
    parsed_mileage, parsed_node_and_conn = parse_mileage_col(mileage), parse_node_col(node)
    parsed_dat = pd.concat([parsed_mileage, parsed_node_and_conn], axis=1, sort=False)
    return parsed_dat


# ====================================================================================================================
class ELRMileages:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = "Engineer's Line References (ELRs)"
        self.URL = self.HomeURL + '/elrs/elr0.shtm'
        self.Catalogue = get_catalogue(self.URL)
        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("line-data", "elrs-and-mileages")
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "dat\\line-data\\elrs-and-mileages" and sub-directories
    def cd_em(self, *sub_dir):
        """
        :param sub_dir: [str]
        :return: [str]
        """
        path = self.DataDir
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\line-data\\elrs-and-mileages\\dat" and sub-directories
    def cdd_em(self, *sub_dir):
        """
        :param sub_dir: [str]
        :return: [str]
        """
        path = self.cd_em("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Scrape Engineer's Line References (ELRs)
    def collect_elr_by_initial(self, initial, update=False, verbose=False):
        """
        :param initial: [str] initial letter of ELR, e.g. 'a', ..., 'z'
        :param update: [bool] whether to re-collect the data
        :param verbose: [bool]
        :return: [dict] {'initial': [pandas.DataFrame], 'Last_updated_date': [str]}
                    [pandas.DataFrame] data of ELRs whose names start with the given 'initial', incl. ELR names,
                    line name, mileages, datum and some notes
                    [str] date of when the data was last updated

        Testing e.g.
            initial = 'a'
            update  = False
            verbose = False
        """
        assert initial in string.ascii_letters
        beginning_with = initial.upper()

        path_to_pickle = self.cd_em("a-z", beginning_with.lower() + ".pickle")
        if os.path.isfile(path_to_pickle) and not update:
            elrs = load_pickle(path_to_pickle)
        else:
            url = self.Catalogue[beginning_with]  # Specify the requested URL
            try:
                source = requests.get(url)  # Request to get connected to the url
                records, header = parse_table(source, parser='lxml')
                # Create a DataFrame of the requested table
                data = pd.DataFrame([[x.replace('=', 'See').strip('\xa0') for x in i] for i in records], columns=header)
                # Return a dictionary containing both the DataFrame and its last updated date
                elrs = {beginning_with: data, 'Last_updated_date': get_last_updated_date(url)}
                save_pickle(elrs, path_to_pickle, verbose)
            except Exception as e:  # e.g the requested URL is not available:
                print("Failed to collect data of ELR beginning with \"{}\". {}".format(beginning_with.upper(), e))
                elrs = {beginning_with: pd.DataFrame(), 'Last_updated_date': ''}
        return elrs

    # Get all ELRs and mileages
    def fetch_elr(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        """
        :param update: [bool] whether to re-collect the data by initial letter
        :param pickle_it: [bool] whether to save the data as a .pickle file
        :param data_dir: [str; None] directory where the data will be stored
        :param verbose: [bool]
        :return [dict] {'ELRs_mileages': [DataFrame], 'Latest_update_date': [str]}
                    [DataFrame] data of (almost all) ELRs beginning with the given 'keyword', including ELR names,
                    line name, mileages, datum and some notes
                    [str] date of when the data was last updated
        """
        data = [self.collect_elr_by_initial(x, update, verbose=False if data_dir or not verbose else True)
                for x in string.ascii_lowercase]
        elrs_data = (item[x] for item, x in zip(data, string.ascii_uppercase))  # Select DataFrames only
        elrs_data_table = pd.concat(elrs_data, axis=0, ignore_index=True, sort=False)

        # Get the latest updated date
        last_updated_dates = (item['Last_updated_date'] for item, _ in zip(data, string.ascii_uppercase))
        latest_update_date = max(d for d in last_updated_dates if d is not None)

        elrs_data = {'ELRs_mileages': elrs_data_table, 'Latest_update_date': latest_update_date}

        if pickle_it and data_dir:
            pickle_filename = "elrs-and-mileages.pickle"
            self.CurrentDataDir = regulate_input_data_dir(data_dir)
            path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
            save_pickle(elrs_data, path_to_pickle, verbose=True)

        return elrs_data

    # Read (from online) the mileage file for the given ELR
    def collect_mileage_file_by_elr(self, elr, parsed=True, confirmation_required=True, pickle_it=False, verbose=False):
        """
        :param elr: [str] e.g. elr='CJD'
        :param parsed: [bool]
        :param confirmation_required: [bool]
        :param pickle_it: [bool]
        :param verbose: [bool]
        :return: [dict]

        Note:
            - In some cases, mileages are unknown hence left blank, e.g. ANI2, Orton Junction with ROB (~3.05)
            - Mileages in parentheses are not on that ELR, but are included for reference,
              e.g. ANL, (8.67) NORTHOLT [London Underground]
            - As with the main ELR list, mileages preceded by a tilde (~) are approximate.

        Testing e.g.
            elr
            parsed = True
            confirmation_required = False
            pickle_it = False
            verbose = False
        """
        if confirmed("To collect mileage file for \"{}\"?".format(elr.upper()),
                     confirmation_required=confirmation_required):

            try:
                # The URL of the mileage file for the ELR
                url = self.HomeURL + '/elrs/_mileages/{}/{}.shtm'.format(elr[0].lower(), elr.lower())
                source = requests.get(url)
                source_text = bs4.BeautifulSoup(source.text, 'lxml')

                line_name, sub_line_name = source_text.find('h3').text, source_text.find('h4')

                if line_name == '"404" error: page not found':
                    initial = elr[0]
                    elr_dat = self.collect_elr_by_initial(initial, verbose=verbose)[initial]
                    elr_dat = elr_dat[elr_dat.ELR == elr]

                    notes = elr_dat.Notes.values[0]
                    if re.match(r'(Now( part of)? |= |See )[A-Z]{3}(\d)?$', notes):
                        new_elr = re.search(r'(?<= )[A-Z]{3}(\d)?', notes).group(0)
                        mileage_file = self.fetch_mileage_file(elr=new_elr, pickle_it=pickle_it)
                        return mileage_file

                    else:
                        line_name, mileages = elr_dat[['Line name', 'Mileages']].values[0]
                        if re.match(r'(\w ?)+ \((\w+ \w+)+.\)', line_name):
                            line_name_ = re.search(r'(?<=\w \()(\w+ \w+)+.(?=\))', line_name).group(0)
                            try:
                                location_a, _, location_b = re.split(r' (and|&|to) ', line_name_)
                                line_name = re.search(r'(\w+ \w+)+.(?= \((\w ?)+\))', line_name).group(0)
                            except ValueError:
                                location_a, _, location_b = re.split(r' (and|&|to) ', notes)
                                line_name = line_name_
                        elif elr_dat.Mileages.values[0].startswith('0.00') and elr_dat.Datum.values[0] != '':
                            location_a = elr_dat.Datum.values[0]
                            location_b = re.split(r' (and|&|to) ', line_name)[
                                2] if location_a in line_name else line_name
                        else:
                            try:
                                location_a, _, location_b = re.split(r' (and|&|to) ', notes)
                            except (ValueError, TypeError):
                                location_a, _, location_b = re.split(r' (and|&|to) ', line_name)
                            else:
                                location_a, location_b = '', ''
                        location_b_ = re.sub(r' Branch| Curve', '', location_b) \
                            if re.match(r'.*( Branch| Curve)$', location_b) else location_b

                        miles_chains, locations = mileages.split(' - '), [location_a, location_b_]
                        parsed_content = [[m, l] for m, l in zip(miles_chains, locations)]

                else:
                    line_name = line_name.split('\t')[1]
                    parsed_content = [x.strip().split('\t', 1)
                                      for x in source_text.find('pre').text.splitlines() if x != '']
                    parsed_content = [[y.replace('  ', ' ').replace('\t', ' ') for y in x]
                                      for x in parsed_content]
                    parsed_content = [[''] + x if (len(x) == 1) & ('Note that' not in x[0]) else x
                                      for x in parsed_content]

                # assert sub_headers[0] == elr
                sub_headers = sub_line_name.text.split('\t')[1] if sub_line_name else ''

                # Make a dict of line information
                line_info = {'ELR': elr, 'Line': line_name, 'Sub-Line': sub_headers}

                # Search for note
                note_temp = min(parsed_content, key=len)
                note = note_temp[0] if len(note_temp) == 1 else ''
                if note:
                    if ' Revised distances are thus:' in note:
                        parsed_content[parsed_content.index(note_temp)] = ['', 'Current measure']
                        note = note.replace(' Revised distances are thus:', '')
                    else:
                        parsed_content.remove(note_temp)

                # Create a table of the mileage data
                mileage_data = pd.DataFrame(parsed_content, columns=['Mileage', 'Node'])

                # Check if there is any missing note
                if mileage_data.iloc[-1].Mileage == '':
                    note = [note, mileage_data.iloc[-1].Node] if note else mileage_data.iloc[-1].Node
                    mileage_data = mileage_data[:-1]

                if len(mileage_data.iloc[-1].Mileage) > 6:
                    note = [note, mileage_data.iloc[-1].Mileage] if note else mileage_data.iloc[-1].Mileage
                    mileage_data = mileage_data[:-1]

                # Make a dict of note
                note_dat = {'Note': note}

                # Identify if there are multiple (both current and former) measures in 'mileage_data'
                mileage_data = identify_multiple_measures(mileage_data)

                if parsed:
                    if isinstance(mileage_data, dict) and len(mileage_data) > 1:
                        mileage_data = {h: parse_mileage_data(dat) for h, dat in mileage_data.items()}
                    else:  # isinstance(dat, pd.DataFrame)
                        mileage_data = parse_mileage_data(mileage_data)

                mileage_file = dict(pair for x in [line_info, {elr: mileage_data}, note_dat] for pair in x.items())

                if pickle_it:
                    path_to_pickle = self.cd_em("mileage-files", elr[0].lower(), elr + ".pickle")
                    if os.path.basename(path_to_pickle) == "prn.pickle":
                        path_to_pickle = path_to_pickle.replace("prn.pickle", "prn_x.pickle")
                    save_pickle(mileage_file, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect the mileage file for \"{}\". {}.".format(elr, e)) if verbose else None
                mileage_file = None

            return mileage_file

    # Get the mileage file for the given ELR (firstly try to load the local data file if available)
    def fetch_mileage_file(self, elr, update=False, pickle_it=False, data_dir=None, verbose=False):
        """
        :param elr: [str]
        :param update: [bool] indicate whether to re-scrape the data from online
        :param pickle_it: [bool]
        :param data_dir: [str; None]
        :param verbose: [bool]
        :return: [dict] {elr: [DataFrame] mileage file data,
                        'Line': [str] line name,
                        'Note': [str] additional information/notes, or None}
        """
        path_to_pickle = self.cd_em("mileage-files", elr[0].lower(), elr + ".pickle")
        if os.path.basename(path_to_pickle) == "prn.pickle":
            path_to_pickle = path_to_pickle.replace("prn.pickle", "prn_x.pickle")

        if os.path.isfile(path_to_pickle) and not update:
            mileage_file = load_pickle(path_to_pickle)

        else:
            mileage_file = self.collect_mileage_file_by_elr(elr, parsed=True, confirmation_required=False,
                                                            pickle_it=pickle_it,
                                                            verbose=False if data_dir or not verbose else True)

            if mileage_file:
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, os.path.basename(path_to_pickle))
                    save_pickle(mileage_file, path_to_pickle, verbose=True)
            else:
                print("No mileage file has been collected for \"{}\".".format(elr.upper())) if verbose else None

        return mileage_file

    @staticmethod
    def search_conn(start_elr, start_em, end_elr, end_em):
        """
        :param start_elr: start_elr
        :param start_em: start_em=conn_em.copy()
        :param end_elr: end_elr=conn_elr
        :param end_em:
        :return:
        """
        start_mask = start_em.apply(lambda x: x.str.contains(end_elr, case=False).any(), axis=1)
        start_temp = start_em[start_mask]
        assert isinstance(start_temp, pd.DataFrame)
        if not start_temp.empty:
            # Get exact location
            key_idx = start_temp.index[0]
            mile_chain_col = [x for x in start_temp.columns if re.match(r'.*_Mile_Chain', x)][0]
            # Mileage of the Start ELR
            start_dest_mileage = start_em.loc[key_idx, 'Mileage']
            # Mileage of the End ELR
            end_orig_mile_chain = start_temp.loc[key_idx, mile_chain_col]
            if end_orig_mile_chain and end_orig_mile_chain != 'Unknown':
                end_orig_mileage = mile_chain_to_nr_mileage(end_orig_mile_chain)
            else:  # end_conn_mile_chain == '':
                end_mask = end_em.apply(lambda x: x.str.contains(start_elr, case=False).any(), axis=1)
                end_temp = end_em[end_mask]
                if not end_temp.empty:
                    end_orig_mileage = end_temp.Mileage.iloc[0]
                else:
                    end_orig_mileage = start_dest_mileage
        else:
            start_dest_mileage, end_orig_mileage = '', ''
        return start_dest_mileage, end_orig_mileage

    # Get to end and start mileages for StartELR and EndELR, respectively, for the connection point
    def get_conn_mileages(self, start_elr, end_elr, update=False, pickle_mileage_file=False, data_dir=None,
                          verbose=False):
        """
        :param start_elr: [str]
        :param end_elr: [str]
        :param update: [bool] (default: False)
        :param pickle_mileage_file: [bool] (default: False)
        :param data_dir: [str; None (default)]
        :param verbose: [bool] (default: False)
        :return: [tuple]

        Testing e.g.
            start_elr, end_elr = 'NAY', 'LTN2'
            update = False
            pickle_mileage_file = False
            data_dir = None
            verbose = True
        """
        start_file = self.fetch_mileage_file(start_elr, update, pickle_mileage_file, data_dir, verbose=verbose)
        end_file = self.fetch_mileage_file(end_elr, update, pickle_mileage_file, data_dir, verbose=verbose)

        if start_file is not None and end_file is not None:
            start_elr, end_elr = start_file['ELR'], end_file['ELR']
            start_em, end_em = start_file[start_elr], end_file[end_elr]
            key_pat = re.compile(r'(Current\s)|(One\s)|(Later\s)|(Usual\s)')
            if isinstance(start_em, dict):
                start_em = start_em[[k for k in start_em.keys() if re.match(key_pat, k)][0]]
            if isinstance(end_em, dict):
                end_em = end_em[[k for k in end_em.keys() if re.match(key_pat, k)][0]]
            #
            start_dest_mileage, end_orig_mileage = self.search_conn(start_elr, start_em, end_elr, end_em)

            conn_elr, conn_orig_mileage, conn_dest_mileage = '', '', ''

            if not start_dest_mileage and not end_orig_mileage:
                link_cols = [x for x in start_em.columns if re.match(r'Link_\d_ELR.?', x)]
                conn_elrs = start_em[link_cols]

                i = 0
                while i < len(link_cols):
                    link_col = link_cols[i]
                    conn_temp = conn_elrs[conn_elrs.astype(bool)].dropna(how='all')[link_col].dropna()

                    j = 0
                    while j < len(conn_temp):
                        conn_elr = conn_temp.iloc[j]
                        conn_em = self.fetch_mileage_file(conn_elr, update=update)
                        if conn_em is not None:
                            conn_elr = conn_em['ELR']
                            conn_em = conn_em[conn_elr]
                            if isinstance(conn_em, dict):
                                conn_em = conn_em[[k for k in conn_em.keys() if re.match(key_pat, k)][0]]
                            #
                            start_dest_mileage, conn_orig_mileage = \
                                self.search_conn(start_elr, start_em, conn_elr, conn_em)
                            #
                            conn_dest_mileage, end_orig_mileage = \
                                self.search_conn(conn_elr, conn_em, end_elr, end_em)

                            if conn_dest_mileage and end_orig_mileage:
                                if not start_dest_mileage:
                                    start_dest_mileage = start_em[start_em[link_col] == conn_elr].Mileage.values[0]
                                if not conn_orig_mileage:
                                    link_col_conn = \
                                        conn_em.where(conn_em == start_elr).dropna(axis=1, how='all').columns[0]
                                    conn_orig_mileage = conn_em[conn_em[link_col_conn] == start_elr].Mileage.values[0]
                                break
                            else:
                                conn_elr = ''
                                j += 1
                        else:
                            j += 1

                    if conn_elr != '':
                        break
                    else:
                        i += 1

            if conn_orig_mileage and not conn_elr:
                start_dest_mileage, conn_orig_mileage = '', ''

        else:
            start_dest_mileage, conn_elr, conn_orig_mileage, conn_dest_mileage, end_orig_mileage = [''] * 5

        return start_dest_mileage, conn_elr, conn_orig_mileage, conn_dest_mileage, end_orig_mileage



"""

Data source: http://www.railwaycodes.org.uk

Section codes for overhead line electrification (OLE) installations
(http://www.railwaycodes.org.uk/electrification/mast_prefix0.shtm)

"""

import copy
import os
import re

import bs4
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.ops import confirmed
from pyhelpers.store import load_pickle

from pyrcs.utils import cd_dat, get_catalogue, get_last_updated_date, parse_tr, save_pickle


class Electrification:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Electrification'
        self.URL = self.HomeURL + '/electrification/mast_prefix0.shtm'
        self.Catalogue = get_catalogue(self.URL)
        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("line-data", 'electrification')
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "dat\\line-data\\electrification" and sub-directories
    def cd_elec(self, *sub_dir):
        path = self.DataDir
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\line-data\\electrification\\dat" and sub-directories
    def cdd_elec(self, *sub_dir):
        path = self.cd_elec("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # National network
    def collect_codes_for_national_network(self, confirmation_required=True, verbose=False):
        if confirmed("To collect section codes for OLE installations: national network?",
                     confirmation_required=confirmation_required):
            title_name = 'National network'
            url = self.Catalogue[title_name]

            try:
                source = requests.get(url)
                soup = bs4.BeautifulSoup(source.text, 'lxml')

                codes_for_ole, h3 = {}, soup.find('h3')
                while h3:
                    header_tag = h3.find_next('table')
                    if header_tag:
                        header = [x.text for x in header_tag.find_all('th')]

                        table = pd.DataFrame(parse_tr(header, header_tag.find_next('table').find_all('tr')),
                                             columns=header)
                        table = table.applymap(
                            lambda x: re.sub(r'\']\)?', ']', re.sub(r'\(?\[\'', '[', x)).replace('\\xa0', ''))
                    else:
                        table = pd.DataFrame((x.text for x in h3.find_all_next('li')), columns=['Unknown_codes'])

                    # Notes
                    notes = {'Notes': None}
                    if h3.find_next_sibling().name == 'p':
                        next_p = h3.find_next('p')
                        if next_p.find_previous('h3') == h3:
                            notes['Notes'] = next_p.text.replace('\xa0', '')

                    note_tag = h3.find_next('h4')
                    if note_tag and note_tag.text == 'Notes':
                        notes_ = dict((x.find('a').get('id').title(), x.text.replace('\xa0', ''))
                                      for x in soup.find('ol') if x != '\n')
                        if notes['Notes'] is None:
                            notes['Notes'] = notes_
                        else:
                            notes['Notes'] = [notes['Notes'], notes_]

                    data_key = re.search(r'(\w ?)+(?=( \((\w ?)+\))?)', h3.text).group(0)
                    codes_for_ole.update({data_key.strip(): {'Codes': table, **notes}})

                    h3 = h3.find_next_sibling('h3')

                source.close()

                last_updated_date = get_last_updated_date(url)
                codes_for_ole.update({'Last_updated_date': last_updated_date})

                path_to_pickle = self.cd_elec(title_name.lower().replace(" ", "-") + ".pickle")
                save_pickle(codes_for_ole, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect section codes for OLE installations: national network. {}".format(e))
                codes_for_ole = None

            return codes_for_ole

    # Fetch codes for national network
    def fetch_codes_for_national_network(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        pickle_filename = "national-network.pickle"
        path_to_pickle = self.cd_elec(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            codes_for_ole = load_pickle(path_to_pickle)

        else:
            codes_for_ole = self.collect_codes_for_national_network(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)

            if codes_for_ole:  # codes_for_ole is not None
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
                    save_pickle(codes_for_ole, path_to_pickle, verbose=True)
            else:
                print("No data of section codes has been collected for national network OLE installations.")

        return codes_for_ole

    # Get names of independent lines
    def get_names_of_independent_lines(self):
        url = self.Catalogue['Independent lines']
        source = requests.get(url)
        soup = bs4.BeautifulSoup(source.text, 'lxml')
        for x in soup.find_all('p'):
            if re.match(r'^Jump to: ', x.text):
                line_names = x.text.replace('Jump to: ', '').split(' | ')
                return line_names

    # Independent lines
    def collect_codes_for_independent_lines(self, confirmation_required=True, verbose=False):
        if confirmed("To collect section codes for OLE installations: independent lines?",
                     confirmation_required=confirmation_required):
            title_name = 'Independent lines'
            url = self.Catalogue[title_name]
            try:
                source = requests.get(url)
                soup = bs4.BeautifulSoup(source.text, 'lxml')

                codes_for_independent_lines = {}
                h3 = soup.find('h3')
                while h3:
                    header_tag, table = h3.find_next('table'), None
                    if header_tag:
                        if header_tag.find_previous('h3') == h3:
                            header = [x.text for x in header_tag.find_all('th')]
                            table = pd.DataFrame(parse_tr(header, header_tag.find_next('table').find_all('tr')),
                                                 columns=header)
                            table = table.applymap(
                                lambda x: re.sub(
                                    r'\']\)?', ']', re.sub(r'\(?\[\'', '[', x)).replace('\\xa0', '').strip())

                    notes = {'Notes': None}
                    h4 = h3.find_next('h4')
                    if h4:
                        previous_h3 = h4.find_previous('h3')
                        if previous_h3 == h3 and h4.text == 'Notes':
                            notes_ = dict((x.find('a').get('id').title(), x.text.replace('\xa0', ''))
                                          for x in h4.find_next('ol') if x != '\n')
                            if notes['Notes'] is None:
                                notes['Notes'] = notes_

                    note_tag, note_txt = h3.find_next('p'), ''
                    if note_tag:
                        previous_h3 = note_tag.find_previous('h3')
                        if previous_h3 == h3:
                            note_txt = note_tag.text.replace('\xa0', '')
                            if notes['Notes'] is None:
                                notes['Notes'] = note_txt
                            else:
                                notes['Notes'] = [notes['Notes'], note_txt]

                    ex_note_tag = note_tag.find_next('ol')
                    if ex_note_tag:
                        previous_h3 = ex_note_tag.find_previous('h3')
                        if previous_h3 == h3:
                            li = dict(re.sub(r'[()]', '', x.text).split(' ', 1) for x in ex_note_tag.find_all('li'))
                            notes.update(li)

                    codes_for_independent_lines.update({h3.text: {'Codes': table, **notes}})

                    h3 = h3.find_next_sibling('h3')

                source.close()

                last_updated_date = get_last_updated_date(url)
                codes_for_independent_lines.update({'Last_updated_date': last_updated_date})

                path_to_pickle = self.cd_elec(title_name.lower().replace(" ", "-") + ".pickle")
                save_pickle(codes_for_independent_lines, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect section codes for OLE installations: independent lines. {}".format(e))
                codes_for_independent_lines = None

            return codes_for_independent_lines

    # Fetch codes for independent lines
    def fetch_codes_for_independent_lines(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        pickle_filename = "independent-lines.pickle"
        path_to_pickle = self.cd_elec(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            codes_for_independent_lines = load_pickle(path_to_pickle)

        else:
            codes_for_independent_lines = self.collect_codes_for_independent_lines(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)

            if codes_for_independent_lines:  # codes_for_independent_lines is not None
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
                    save_pickle(codes_for_independent_lines, path_to_pickle, verbose=True)
            else:
                print("No data of section codes has been collected for independent lines OLE installations.")

        return codes_for_independent_lines

    # National network neutral sections
    def collect_codes_for_ohns(self, confirmation_required=True, verbose=False):
        if confirmed("To collect codes for OLE neutral sections (OHNS)?", confirmation_required=confirmation_required):
            title_name = 'National network neutral sections'
            url = self.Catalogue[title_name]

            try:
                header, neutral_sections_data = pd.read_html(url)
                neutral_sections_data.columns = header.columns.to_list()
                neutral_sections_data.fillna('', inplace=True)

                last_updated_date = get_last_updated_date(url)
                ohns = {'Codes': neutral_sections_data, 'Last_updated_date': last_updated_date}

                path_to_pickle = self.cd_elec(title_name.lower().replace(" ", "-") + ".pickle")
                save_pickle(ohns, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect codes for OLE neutral sections (OHNS). {}".format(e))
                ohns = None

            return ohns

    # Fetch codes for Overhead line electrification neutral sections (OHNS)
    def fetch_codes_for_ohns(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        pickle_filename = "national-network-neutral-sections.pickle"
        path_to_pickle = self.cd_elec(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            ohns = load_pickle(path_to_pickle)

        else:
            ohns = self.collect_codes_for_ohns(confirmation_required=False,
                                               verbose=False if data_dir or not verbose else True)

            if ohns:  # ohns is not None
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
                    save_pickle(ohns, path_to_pickle, verbose=True)
            else:
                print("No data of section codes for OHNS has been collected.")

        return ohns

    # National network energy tariff zones
    def collect_codes_for_energy_tariff_zones(self, confirmation_required=True, verbose=False):
        if confirmed("To collect codes for the UK railway electrification tariff zones?",
                     confirmation_required=confirmation_required):
            title_name = 'National network energy tariff zones'
            url = self.Catalogue[title_name]
            try:
                source = requests.get(url)
                soup = bs4.BeautifulSoup(source.text, 'lxml')

                codes_for_energy_tariff_zones = {}
                h3 = soup.find('h3')
                while h3:
                    header_tag, table = h3.find_next('table'), None
                    if header_tag:
                        if header_tag.find_previous('h3') == h3:
                            header = [x.text for x in header_tag.find_all('th')]
                            table = pd.DataFrame(parse_tr(header, header_tag.find_next('table').find_all('tr')),
                                                 columns=header)
                            table = table.applymap(
                                lambda x: re.sub(r'\']\)?', ']', re.sub(r'\(?\[\'', '[', x)).replace(
                                    '\\xa0', '').strip())

                    notes, next_p = [], h3.find_next('p')
                    previous_h3 = next_p.find_previous('h3')
                    while previous_h3 == h3:
                        notes.append(next_p.text.replace('\xa0', ''))
                        next_p = next_p.find_next('p')
                        try:
                            previous_h3 = next_p.find_previous('h3')
                        except AttributeError:
                            break
                    notes = ' '.join(notes).strip()

                    codes_for_energy_tariff_zones.update({h3.text: {'Codes': table, 'Notes': notes}})

                    h3 = h3.find_next_sibling('h3')

                source.close()

                last_updated_date = get_last_updated_date(url)
                codes_for_energy_tariff_zones.update({'Last_updated_date': last_updated_date})

                path_to_pickle = self.cd_elec(title_name.lower().replace(" ", "-") + ".pickle")
                save_pickle(codes_for_energy_tariff_zones, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect the codes for UK railway electrification tariff zones. {}".format(e))
                codes_for_energy_tariff_zones = None

            return codes_for_energy_tariff_zones

    # Fetch codes for Overhead line electrification neutral sections (OHNS)
    def fetch_codes_for_energy_tariff_zones(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        pickle_filename = "national-network-energy-tariff-zones.pickle"
        path_to_pickle = self.cd_elec(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            codes_for_energy_tariff_zones = load_pickle(path_to_pickle)

        else:
            codes_for_energy_tariff_zones = self.collect_codes_for_energy_tariff_zones(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)

            if codes_for_energy_tariff_zones:  # codes_for_energy_tariff_zones is not None
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
                    save_pickle(codes_for_energy_tariff_zones, path_to_pickle, verbose=True)
            else:
                print("No data of section codes has been collected for the UK railway electrification tariff zones.")

        return codes_for_energy_tariff_zones

    # Fetch codes in the electrification catalogue
    def fetch_electrification_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        national_network = self.fetch_codes_for_national_network(update, verbose=verbose)
        independent_lines = self.fetch_codes_for_independent_lines(update, verbose=verbose)
        ohns = self.fetch_codes_for_ohns(update, verbose=verbose)
        energy_tariff_zones = self.fetch_codes_for_energy_tariff_zones(update, verbose=verbose)

        items = list(self.Catalogue.keys())
        items.remove('Introduction')

        electrification_codes = dict(zip(items, [national_network, independent_lines, ohns, energy_tariff_zones]))

        if pickle_it and data_dir:
            pickle_filename = "electrification-codes.pickle"
            self.CurrentDataDir = regulate_input_data_dir(data_dir)
            path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
            save_pickle(electrification_codes, path_to_pickle, verbose=True)

        return electrification_codes



""" Line data """

from pyrcs.line_data_cls import crs_nlc_tiploc_stanox, electrification, elrs_mileages, lor_codes
from pyrcs.line_data_cls import line_names, track_diagrams
from pyrcs.utils import get_cls_menu


class LineData:
    def __init__(self):
        # Basic info
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Line data'
        self.URL = self.HomeURL + '/linedatamenu.shtm'
        self.Catalogue = get_cls_menu(self.URL)
        # Classes
        self.ELRMileages = elrs_mileages.ELRMileages()
        self.Electrification = electrification.Electrification()
        self.LocationIdentifiers = crs_nlc_tiploc_stanox.LocationIdentifiers()
        self.LOR = lor_codes.LOR()
        self.LineNames = line_names.LineNames()
        self.TrackDiagrams = track_diagrams.TrackDiagrams()



""" Utilities - Helper functions """

import collections
import datetime
import numbers
import os
import pickle
import re
import urllib.parse

import bs4
import dateutil.parser
import measurement.measures
import numpy as np
import pandas as pd
import pkg_resources
import pyhelpers.ops
import pyhelpers.store
import rapidjson
import requests

# ====================================================================================================================
""" Change directory """


# Change directory to "dat" and sub-directories
def cd_dat(*sub_dir, dat_dir="dat", mkdir=False) -> str:
    """
    :param sub_dir: [str]
    :param dat_dir: [str] (default: "dat")
    :param mkdir: [bool] (default: False)
    :return: [str]
    """
    path = pkg_resources.resource_filename(__name__, dat_dir)
    for x in sub_dir:
        path = os.path.join(path, x)
    if mkdir:
        os.makedirs(path, exist_ok=True)
    return path


# ====================================================================================================================
""" Save data """


# Save Pickle file
def save_pickle(pickle_data, path_to_pickle, verbose=True):
    """
    :param pickle_data: any object that could be dumped by the 'pickle' package
    :param path_to_pickle: [str] local file path
    :param verbose: [bool] (default: True)
    :return: whether the data has been successfully saved
    """
    pickle_filename = os.path.basename(path_to_pickle)
    pickle_dir = os.path.basename(os.path.dirname(path_to_pickle))
    pickle_dir_parent = os.path.basename(os.path.dirname(os.path.dirname(path_to_pickle)))

    if verbose:
        print("{} \"{}\" ... ".format("Updating" if os.path.isfile(path_to_pickle) else "Saving",
                                      " - ".join([pickle_dir_parent, pickle_dir, pickle_filename])), end="")

    try:
        os.makedirs(os.path.dirname(os.path.abspath(path_to_pickle)), exist_ok=True)
        pickle_out = open(path_to_pickle, 'wb')
        pickle.dump(pickle_data, pickle_out)
        pickle_out.close()
        print("Successfully.") if verbose else None
    except Exception as e:
        print("Failed. {}.".format(e))


# Save JSON file
def save_json(json_data, path_to_json, verbose=True):
    """
    :param json_data: any object that could be dumped by the 'json' package
    :param path_to_json: [str] local file path
    :param verbose: [bool] (default: True)
    :return: whether the data has been successfully saved
    """
    json_filename = os.path.basename(path_to_json)
    json_dir = os.path.basename(os.path.dirname(path_to_json))
    json_dir_parent = os.path.basename(os.path.dirname(os.path.dirname(path_to_json)))

    print("{} \"{}\" ... ".format("Updating" if os.path.isfile(path_to_json) else "Saving",
                                  " - ".join([json_dir_parent, json_dir, json_filename])), end="") if verbose else None
    try:
        os.makedirs(os.path.dirname(os.path.abspath(path_to_json)), exist_ok=True)
        json_out = open(path_to_json, 'w')
        rapidjson.dump(json_data, json_out)
        json_out.close()
        print("Successfully.") if verbose else None
    except Exception as e:
        print("Failed. {}.".format(e))


# ====================================================================================================================
""" Converter """


# Convert "miles.chains" to Network Rail mileages
def mile_chain_to_nr_mileage(miles_chains) -> str:
    """
    :param miles_chains: [str; np.nan; None] 'miles.chains'
    :return: [str] 'miles.yards'

    Note on the 'ELRs and mileages' web page that 'mileages' are given in the form 'miles.chains'.

    Testing e.g.
        miles_chains = '0.18'  # AAM 0.18 Tewkesbury Junction with ANZ (84.62)
        mile_chain_to_nr_mileage(miles_chains)  # '0.0396'
        miles_chains = ''  # np.nan  # None
        mile_chain_to_nr_mileage(miles_chains)  # ''
    """
    if pd.notna(miles_chains) and miles_chains != '':
        miles, chains = str(miles_chains).split('.')
        yards = measurement.measures.Distance(chain=chains).yd
        network_rail_mileage = '%.4f' % (int(miles) + round(yards / (10 ** 4), 4))
    else:
        network_rail_mileage = ''
    return network_rail_mileage


# Convert Network Rail mileages to "miles.chains"
def nr_mileage_to_mile_chain(str_mileage) -> str:
    """
    :param str_mileage: [str; np.nan; None] 'miles.yards'
    :return: [str] 'miles.chains'

    Note on the 'ELRs and mileages' web page that 'mileages' are given in the form 'miles.chains'.

    Testing e.g.
        str_mileage = '0.0396'
        nr_mileage_to_mile_chain(str_mileage)  # '0.18'
        str_mileage = ''  # np.nan  # None
        nr_mileage_to_mile_chain(str_mileage)  # ''
    """
    if pd.notna(str_mileage) and str_mileage != '':
        miles, yards = str(str_mileage).split('.')
        chains = measurement.measures.Distance(yard=yards).chain
        miles_chains = '%.2f' % (int(miles) + round(chains / (10 ** 2), 2))
    else:
        miles_chains = ''
    return miles_chains


# Convert str type Network Rail mileage to numerical type
def nr_mileage_str_to_num(str_mileage: str) -> float:
    """
    Testing e.g.
        str_mileage = '0.0396'
        nr_mileage_str_to_num(str_mileage)  # 0.0396
        str_mileage = ''
        nr_mileage_str_to_num(str_mileage)  # np.nan
    """
    num_mileage = np.nan if str_mileage == '' else round(float(str_mileage), 4)
    return num_mileage


# Convert Network Rail mileage to str type
def nr_mileage_num_to_str(num_mileage: float) -> str:
    """
    Testing e.g.
        num_mileage = 0.0396
        nr_mileage_num_to_str(num_mileage)  # '0.0396'
        num_mileage = np.nan
        nr_mileage_num_to_str(num_mileage)  # ''
    """
    nr_mileage = '%.4f' % round(float(num_mileage), 4) if num_mileage and pd.notna(num_mileage) else ''
    return nr_mileage


# Convert Network Rail mileages to yards
def nr_mileage_to_yards(nr_mileage: (float, str)) -> int:
    """
    Testing e.g.
        nr_mileage = '0.0396'
        nr_mileage_to_yards(nr_mileage)  # 396
        nr_mileage = 0.0396
        nr_mileage_to_yards(nr_mileage)  # 396
    """
    if isinstance(nr_mileage, (float, np.float, int, np.integer)):
        nr_mileage = nr_mileage_num_to_str(nr_mileage)
    else:
        pass
    miles = int(nr_mileage.split('.')[0])
    yards = int(nr_mileage.split('.')[1])
    yards += int(measurement.measures.Distance(mi=miles).yd)
    return yards


# Convert yards to Network Rail mileages
def yards_to_nr_mileage(yards: (int, float, np.nan)) -> str:
    """
    Testing e.g.
        yards = 396
        yards_to_nr_mileage(yards)  # '0.0396'
        yards = 396.0
        yards_to_nr_mileage(yards)  # '0.0396'
        yards = None  # np.nan
        yards_to_nr_mileage(yards)  # ''
    """
    if pd.notnull(yards) and yards != '':
        mileage_mi = np.floor(measurement.measures.Distance(yd=yards).mi)
        mileage_yd = yards - int(measurement.measures.Distance(mi=mileage_mi).yd)
        # Example: "%.2f" % round(2606.89579999999, 2)
        mileage = str('%.4f' % round((mileage_mi + mileage_yd / (10 ** 4)), 4))
    else:
        mileage = ''
    return mileage


# For a location x where (start_mileage_num == end_mileage_num), consider a section [x - shift_yards, x + shift_yards]
def shift_num_nr_mileage(nr_mileage: (float, int, str), shift_yards: (int, float)) -> float:
    """
    :param nr_mileage: [float]
    :param shift_yards: [int]
    :return: [float]

    Testing e.g.
        nr_mileage  = '0.0396'  # 0.0396  # 10
        shift_yards = 220  # 220.99
        shift_num_mileage(nr_mileage, shift_yards)  # 0.0616  # 0.0617
    """
    yards = nr_mileage_to_yards(nr_mileage) + shift_yards
    shifted_nr_mileage = yards_to_nr_mileage(yards)
    shifted_num_mileage = nr_mileage_str_to_num(shifted_nr_mileage)
    return shifted_num_mileage


# Convert calendar year to Network Rail financial year
def year_to_financial_year(date: datetime.datetime) -> int:
    """
    Testing e.g.
        date = datetime.datetime.now()
        year_to_financial_year(date)
    """
    financial_date = date + pd.DateOffset(months=-3)
    return financial_date.year


# ====================================================================================================================
""" Parser """


# Get a list of parsed HTML tr's
def parse_tr(header, trs) -> list:
    """
    :param header: [list] list of column names of a requested table
    :param trs: [bs4.ResultSet - list of bs4.Tag] contents under 'tr' tags of the web page
    :return: [list] list of lists each comprising a row of the requested table

    Get a list of parsed contents of tr-tag's, each of which corresponds to a piece of record
    Reference: https://stackoverflow.com/questions/28763891/

    Testing e.g.
        source = requests.get('http://www.railwaycodes.org.uk/tunnels/tunnels1.shtm')
        parsed_text = bs4.BeautifulSoup(source.text, 'lxml')
        header = [x.text for x in parsed_text.find_all('th')]  # Column names
        trs = parsed_text.find_all('table', attrs={'width': '1100px'})[1].find_all('tr')
        parse_tr(header, trs)
    """
    tbl_lst = []
    for row in trs:
        data = []
        for dat in row.find_all('td'):
            txt = dat.get_text()
            if '\n' in txt:
                t = txt.split('\n')
                txt = '%s (%s)' % (t[0], t[1:]) if '(' not in txt and ')' not in txt else '%s %s' % (t[0], t[1:])
                data.append(txt)
            else:
                data.append(txt)
        tbl_lst.append(data)

    row_spanned = []
    for no, tr in enumerate(trs):
        for td_no, rho in enumerate(tr.find_all('td')):
            # print(data.has_attr("rowspan"))
            if rho.has_attr('rowspan'):
                row_spanned.append((no, int(rho['rowspan']), td_no, rho.text))

    if row_spanned:
        d = collections.defaultdict(list)
        for k, *v in row_spanned:
            d[k].append(v)
        row_spanned = list(d.items())

        for x in row_spanned:
            i, to_repeat = x[0], x[1]
            for y in to_repeat:
                for j in range(1, y[0]):
                    if y[2] in tbl_lst[i] and y[2] != '\xa0':
                        y[1] += np.abs(tbl_lst[i].index(y[2]) - y[1])
                    tbl_lst[i + j].insert(y[1], y[2])

    # if row_spanned:
    #     for x in row_spanned:
    #         for j in range(1, x[2]):
    #             # Add value in next tr
    #             idx = x[0] + j
    #             # assert isinstance(idx, int)
    #             if x[1] >= len(tbl_lst[idx]):
    #                 tbl_lst[idx].insert(x[1], x[3])
    #             elif x[3] in tbl_lst[x[0]]:
    #                 tbl_lst[idx].insert(tbl_lst[x[0]].index(x[3]), x[3])
    #             else:
    #                 tbl_lst[idx].insert(x[1] + 1, x[3])

    for k in range(len(tbl_lst)):
        n = len(header) - len(tbl_lst[k])
        if n > 0:
            tbl_lst[k].extend(['\xa0'] * n)
        elif n < 0 and tbl_lst[k][2] == '\xa0':
            del tbl_lst[k][2]

    return tbl_lst


# Parse the acquired list to make it be ready for creating the DataFrame
def parse_table(source, parser='lxml') -> tuple:
    """
    :param source: [requests.Response] response object to connecting a URL to request a table
    :param parser: [str] (default: 'lxml'; alternatives: 'html5lib', 'html.parser')
    :return [tuple] ([list] of lists each comprising a row of the requested table - (see also parse_trs())
                     [list] of column names of the requested table)

    Testing e.g.
        source = requests.get('http://www.railwaycodes.org.uk/tunnels/tunnels1.shtm')
        parser = 'lxml'
        parse_table(source, parser='lxml')
    """
    # Get plain text from the source URL
    web_page_text = source.text  # (If source.status_code == 200, the requested URL is available.)
    # Parse the text
    parsed_text = bs4.BeautifulSoup(web_page_text, parser)
    # Get all data under the HTML label 'tr'
    table_temp = parsed_text.find_all('tr')
    # Get a list of column names for output DataFrame
    headers = table_temp[0]
    header = [header.text for header in headers.find_all('th')]
    # Get a list of lists, each of which corresponds to a piece of record
    trs = table_temp[1:]
    # Return a list of parsed tr's, each of which corresponds to one df row
    return parse_tr(header, trs), header


# Parse location note
def parse_location_note(location_dat) -> tuple:
    """
    :param location_dat: [str; None]
    :return: [tuple] ([str] - Location name, [str] - Note)

    Testing e.g.
        location_dat = 'Abbey Wood'
        parse_location_note(location_dat)
        location_dat = 'Abercynon (formerly Abercynon South)'
        parse_location_note(location_dat)
        location_dat = 'Allerton (reopened as Liverpool South Parkway)'
        parse_location_note(location_dat)
        location_dat = 'Ashford International [domestic portion]'
        parse_location_note(location_dat)
    """
    # Location name
    d = re.search(r'.*(?= \[[\"\']\()', location_dat)
    if d is not None:
        dat = d.group()
    elif ' [unknown feature, labelled "do not use"]' in location_dat:
        dat = re.search(r'\w+(?= \[unknown feature, )', location_dat).group()
    elif ') [formerly' in location_dat:
        dat = re.search(r'.*(?= \[formerly)', location_dat).group()
    else:
        m_pattern = re.compile(
            r'[Oo]riginally |[Ff]ormerly |[Ll]ater |[Pp]resumed | \(was | \(in | \(at | \(also |'
            r' \(second code |\?|\n| \(\[\'| \(definition unknown\)| \(reopened |( portion])$')
        x_tmp = re.search(r'(?=[\[(]).*(?<=[\])])|(?=\().*(?<=\) \[)', location_dat)
        x_tmp = x_tmp.group() if x_tmp is not None else location_dat
        dat = ' '.join(location_dat.replace(x_tmp, '').split()) if re.search(m_pattern, location_dat) else location_dat

    # Note
    y = location_dat.replace(dat, '', 1).strip()
    if y == '':
        note = ''
    else:
        n = re.search(r'(?<=[\[(])[\w ,?]+(?=[])])', y)
        if n is None:
            n = re.search(r'(?<=(\[[\'\"]\()|(\([\'\"]\[)|(\) \[)).*(?=(\)[\'\"]\])|(\][\'\"]\))|\])', y)
        elif '"now deleted"' in y and y.startswith('(') and y.endswith(')'):
            n = re.search(r'(?<=\().*(?=\))', y)
        note = n.group() if n is not None else ''
        if note.endswith('\'') or note.endswith('"'):
            note = note[:-1]

    if 'STANOX ' in dat and 'STANOX ' in location_dat and note == '':
        dat = location_dat[0:location_dat.find('STANOX')].strip()
        note = location_dat[location_dat.find('STANOX'):]

    return dat, note


# Parse date string
def parse_date(str_date, as_date_type=False) -> (str, datetime.date):
    """
    :param str_date: [str]
    :param as_date_type: [bool] (default: False)
    :return: [str; datetime.date] the date formatted as needed

    Testing e.g.
        str_date     = '2019-01-01'
        as_date_type = True
        parse_date(str_date, as_date_type)
    """
    temp_date = dateutil.parser.parse(str_date, fuzzy=True)  # datetime.strptime(last_update_date[12:], '%d %B %Y')
    parsed_date = temp_date.date() if as_date_type else str(temp_date.date())
    return parsed_date


# ====================================================================================================================
""" Get useful information """


# Get last update date
def get_last_updated_date(url, parsed=True, date_type=False) -> (str, None):
    """
    :param url: [str] URL link of a requested web page
    :param parsed: [bool] (default: True) indicator of whether to reformat the date
    :param date_type: [bool] (default: False)
    :return:[str; None] date of when the specified web page was last updated

    Testing e.g.
        url       = 'http://www.railwaycodes.org.uk/crs/CRSa.shtm'
        url_      = 'http://www.railwaycodes.org.uk/linedatamenu.shtm'
        parsed    = True
        date_type = False
        get_last_updated_date(url, parsed, date_type)
        get_last_updated_date(url_, parsed, date_type)  # None
    """
    # Request to get connected to the given url
    source = requests.get(url)
    web_page_text = source.text
    # Parse the text scraped from the requested web page
    parsed_text = bs4.BeautifulSoup(web_page_text, 'lxml')  # (Alternative parsers: 'html5lib', 'html.parser')
    # Find 'Last update date'
    update_tag = parsed_text.find('p', {'class': 'update'})
    if update_tag is not None:
        last_update_date = update_tag.text
        # Decide whether to convert the date's format
        if parsed:
            # Convert the date to "yyyy-mm-dd" format
            last_update_date = parse_date(last_update_date, date_type)
    else:
        last_update_date = None  # print('Information not available.')
    return last_update_date


# Get the catalogue for a class
def get_catalogue(cls_url, navigation_bar_exists=True, menu_exists=True) -> dict:
    """
    :param cls_url: [str]
    :param navigation_bar_exists: [bool] (default: True)
    :param menu_exists: [bool] (default: True)
    :return: [dict] {[str] - title: [str] - URL}

    Testing e.g.
        url                   = 'http://www.railwaycodes.org.uk/crs/CRS0.shtm'
        navigation_bar_exists = True
        menu_exists           = True
        get_catalogue(url, navigation_bar_exists, menu_exists)
    """
    source = requests.get(cls_url)

    if navigation_bar_exists:
        cold_soup = bs4.BeautifulSoup(source.text, 'lxml').find_all('a', text=True, attrs={'class': None})
        lst_reversed = list(reversed(cold_soup))
        hot_soup = []
        for x in lst_reversed:
            if x.text == 'Introduction':
                starting_idx = len(cold_soup) - lst_reversed.index(x) - 1
                hot_soup = cold_soup[starting_idx:]
                break
    else:
        hot_soup = bs4.BeautifulSoup(source.text, 'lxml').find_all('a', text=True, attrs={'class': None})[-6:] \
            if menu_exists else []

    source.close()

    raw_contents = [{x.text: urllib.parse.urljoin(os.path.dirname(cls_url) + '/', x['href'])} for x in hot_soup]

    contents = dict(e for d in raw_contents for e in d.items())

    return contents


# Get a menu of the available classes
def get_cls_menu(cls_url: str) -> dict:
    """
    Testing e.g.
        cls_url = 'http://www.railwaycodes.org.uk/linedatamenu.shtm'
        get_cls_menu(cls_url)
    """
    source = requests.get(cls_url)

    soup = bs4.BeautifulSoup(source.text, 'lxml')
    h1, h2s = soup.find('h1'), soup.find_all('h2')

    cls_name = h1.text.replace(' menu', '')

    if len(h2s) == 0:
        cls_elem = dict((x.text, urllib.parse.urljoin(cls_url, x.get('href'))) for x in h1.find_all_next('a'))

    else:
        all_next = [x.replace(':', '') for x in h1.find_all_next(string=True) if x != '\n' and x != '\xa0'][2:]
        h2s_list = [x.text.replace(':', '') for x in h2s]
        all_next_a = [(x.text, urllib.parse.urljoin(cls_url, x.get('href'))) for x in h1.find_all_next('a', href=True)]

        idx = [all_next.index(x) for x in h2s_list]
        for i in idx:
            all_next_a.insert(i, all_next[i])

        cls_elem, i = {}, 0
        while i <= len(idx):
            if i == 0:
                d = dict(all_next_a[i:idx[i]])
            elif i < len(idx):
                d = {h2s_list[i-1]: dict(all_next_a[idx[i-1]+1:idx[i]])}
            else:
                d = {h2s_list[i-1]: dict(all_next_a[idx[i-1]+1:])}
            i += 1
            cls_elem.update(d)

    cls_menu = {cls_name: cls_elem}
    return cls_menu


# ====================================================================================================================
""" Rectification of location names """


# Create a dict for replace location names
def fetch_location_names_repl_dict(k=None, regex=False, as_dataframe=False) -> dict:
    """
    :param k: [str; None (default)]
    :param regex: [bool] (default: False)
    :param as_dataframe: [bool] (default: False)
    :return: [dict]

    Testing e.g.
        k            = None
        regex        = False
        as_dataframe = True
        fetch_location_names_repl_dict(k, regex, as_dataframe)
    """
    json_filename = "location-names-repl{}.json".format("" if not regex else "-regex")
    location_name_repl_dict = pyhelpers.store.load_json(cd_dat(json_filename))

    if regex:
        location_name_repl_dict = {re.compile(k): v for k, v in location_name_repl_dict.items()}

    replacement_dict = {k: location_name_repl_dict} if k else location_name_repl_dict

    if as_dataframe:
        replacement_dict = pd.DataFrame.from_dict(replacement_dict)

    return replacement_dict


# Rectify location names
def update_location_name_repl_dict(new_items, regex, verbose=False):
    """
    :param new_items: [dict]
    :param regex: [bool]
    :param verbose: [bool] (default: False)

    Testing e.g.
        new_items = {}
        regex     = False
        verbose   = True
        update_location_name_repl_dict(new_items, regex, verbose)
    """
    json_filename = "location-names-repl{}.json".format("" if not regex else "-regex")

    new_items_keys = list(new_items.keys())

    if pyhelpers.ops.confirmed("To update \"{}\" with {{\"{}\"... }}?".format(json_filename, new_items_keys[0])):
        path_to_json = cd_dat(json_filename)
        location_name_repl_dict = pyhelpers.store.load_json(path_to_json)

        if any(isinstance(k, re.Pattern) for k in new_items_keys):
            new_items = {k.pattern: v for k, v in new_items.items() if isinstance(k, re.Pattern)}

        location_name_repl_dict.update(new_items)

        save_json(location_name_repl_dict, path_to_json, verbose=verbose)


# ====================================================================================================================
""" Fixers """


# Fix 'STANOX' if it is loaded as numbers
def fix_num_stanox(stanox_x):

    if isinstance(stanox_x, numbers.Number):
        stanox_x = '' if pd.isna(stanox_x) else str(int(stanox_x))

    if len(stanox_x) < 5 and stanox_x != '':
        stanox_x = '0' * (5 - len(stanox_x)) + stanox_x

    return stanox_x



""" Other assets """

from pyrcs.other_assets_cls import depots, features, signal_boxes, stations, tunnels, viaducts
from pyrcs.utils import get_cls_menu


class OtherAssets:
    def __init__(self):
        # Basic info
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Other assets'
        self.URL = self.HomeURL + '/otherassetsmenu.shtm'
        self.Catalogue = get_cls_menu(self.URL)
        # Classes
        self.SignalBoxes = signal_boxes.SignalBoxes()
        self.Tunnels = tunnels.Tunnels()
        self.Viaducts = viaducts.Viaducts()
        self.Stations = stations.Stations()
        self.Depots = depots.Depots()
        self.Features = features.Features()



"""

Data source: http://www.railwaycodes.org.uk

Railway viaducts (http://www.railwaycodes.org.uk/tunnels/tunnels0.shtm)

"""

import copy
import os
import re

import bs4
import fuzzywuzzy.process
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.store import load_pickle

from pyrcs.utils import cd_dat, get_catalogue, get_last_updated_date
from pyrcs.utils import save_pickle


class Viaducts:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Viaducts'
        self.URL = self.HomeURL + '/viaducts/viaducts0.shtm'
        self.Catalogue = get_catalogue(self.URL)
        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("other-assets", "viaducts")
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "dat\\other-assets\\viaducts\\" and sub-directories
    def cd_viaducts(self, *sub_dir):
        path = self.DataDir
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\other-assets\\viaducts\\dat" and sub-directories
    def cdd_viaducts(self, *sub_dir):
        path = self.cd_viaducts("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Get the titles for each page
    def get_page_titles(self):
        intro_page = requests.get(self.URL)
        pages = [x.text for x in bs4.BeautifulSoup(intro_page.text, 'lxml').find_all('a', text=re.compile('^Page.*'))]
        pages = pages[0:int(len(pages) / 2)]
        return pages

    # Collect viaducts data for a given page number
    def collect_railway_viaducts(self, page_no, update=False, verbose=False):
        """
        :param page_no: [int] page number; valid values include 1, 2, 3, 4, 5, and 6
        :param update: [bool]
        :param verbose: [bool]
        :return: [dict]
        """
        assert page_no in range(1, 7), "Valid \"page_no\" must be one of 1, 2, 3, 4, 5, and 6."
        page_headers = self.get_page_titles()
        filename = fuzzywuzzy.process.extractOne(str(page_no), page_headers)[0]
        pickle_filename = re.sub(r"[()]", "", re.sub(r"[ -]", "-", filename)).lower() + ".pickle"

        path_to_pickle = self.cd_viaducts(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            viaducts_codes = load_pickle(path_to_pickle)

        else:
            url = self.URL.replace('viaducts0', 'viaducts{}'.format(page_no))

            try:
                last_updated_date = get_last_updated_date(url)
            except Exception as e:
                print("Failed to find the last updated date for viaducts codes on \"{}\". {}".format(filename, e))
                last_updated_date = ''

            try:
                header, viaducts_table = pd.read_html(url, na_values=[''], keep_default_na=False)
                viaducts_table.columns = header.columns.to_list()
                viaducts_table.fillna('', inplace=True)
            except Exception as e:
                print("Failed to collect viaducts data for Page \"{}\". {}".format(page_no, e))
                viaducts_table = pd.DataFrame()

            viaducts_codes = {re.search(r'(?<=\()\w.*(?=\))', filename).group(0).replace('-', '_'): viaducts_table}
            viaducts_codes.update({'Last_updated_date': last_updated_date})

            save_pickle(viaducts_codes, path_to_pickle, verbose)

        return viaducts_codes

    # Fetch all of the collected viaducts data
    def fetch_railway_viaducts(self, update=False, pickle_it=False, data_dir=None, verbose=False):

        data_sets = [self.collect_railway_viaducts(page_no, update, verbose=False if data_dir or not verbose else True)
                     for page_no in range(1, 7)]

        viaducts_codes_tables = [dat[k] for dat in data_sets for k, v in dat.items() if k != 'Last_updated_date']
        viaducts_codes = pd.concat(viaducts_codes_tables, axis=0, ignore_index=True, sort=False)

        last_updated_dates = [dat[k] for dat in data_sets for k, v in dat.items() if k == 'Last_updated_date']
        latest_update_date = max(last_updated_dates)

        viaducts_codes = {'Viaducts': viaducts_codes, 'Latest_update_date': latest_update_date}

        if pickle_it and data_dir:
            self.CurrentDataDir = regulate_input_data_dir(data_dir)
            path_to_pickle = os.path.join(self.CurrentDataDir, "railway-viaducts.pickle")
            save_pickle(viaducts_codes, path_to_pickle, verbose=True)

        return viaducts_codes



""" Railway tunnel lengths """

import copy
import itertools
import operator
import os
import re

import bs4
import fuzzywuzzy.process
import measurement.measures
import numpy as np
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.store import load_pickle

from pyrcs.utils import cd_dat, get_catalogue, get_last_updated_date, parse_tr
from pyrcs.utils import save_pickle


class Tunnels:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Tunnels'
        self.URL = self.HomeURL + '/tunnels/tunnels0.shtm'
        self.Catalogue = get_catalogue(self.URL)
        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("other-assets", 'tunnels')
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "dat\\other-assets\\tunnels" and sub-directories
    def cd_tunnels(self, *directories):
        path = self.DataDir
        for x in directories:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\other-assets\\tunnels\\dat" and sub-directories
    def cdd_tunnels(self, *sub_dir):
        path = self.cd_tunnels("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Find page headers
    def find_page_headers(self):
        intro_page = requests.get(self.URL)
        pages = [x.text for x in bs4.BeautifulSoup(intro_page.text, 'lxml').find_all('a', text=re.compile('^Page.*'))]
        return pages[:int(len(pages) / 2)]

    # Parse data in 'Length' column, i.e. convert miles/yards to metres
    @staticmethod
    def parse_tunnel_length(x):
        """
        :param x:
        :return:

        Examples:
            '' or Unknown
            1m 182y; 0m111y; c0m 150y; 0m 253y without avalanche shelters
            0m 56ch
            formerly 0m236y
            0.325km (0m 356y); 0.060km ['(0m 66y)']
            0m 48yd- (['0m 58yd'])
        """
        if re.match(r'[Uu]nknown', x):
            length = np.nan
            add_info = 'Unknown'
        elif x == '':
            length = np.nan
            add_info = 'Unavailable'
        elif re.match(r'\d+m \d+yd-.*\d+m \d+yd.*', x):
            miles_a, yards_a, miles_b, yards_b = re.findall(r'\d+', x)
            length_a = measurement.measures.Distance(mi=miles_a).m + measurement.measures.Distance(yd=yards_a).m
            length_b = measurement.measures.Distance(mi=miles_b).m + measurement.measures.Distance(yd=yards_b).m
            length = (length_a + length_b) / 2
            add_info = '-'.join([str(round(length_a, 2)), str(round(length_b, 2))]) + ' metres'
        else:
            if re.match(r'(formerly )?c?\d+m ?\d+y?(ch)?.*', x):
                miles, yards = re.findall(r'\d+', x)
                if re.match(r'.*\d+ch$', x):
                    yards = measurement.measures.Distance(chain=yards).yd
                if re.match(r'^c.*', x):
                    add_info = 'Approximate'
                elif re.match(r'\d+y$', x):
                    add_info = re.search(r'(?<=\dy).*$', x).group()
                elif re.match(r'^(formerly).*', x):
                    add_info = 'Formerly'
                else:
                    add_info = None
            elif re.match(r'\d+\.\d+km .*\(\d+m \d+y\).*', x):
                miles, yards = re.findall(r'\d+', re.search(r'(?<=\()\d+.*(?=\))', x).group())
                add_info = re.search(r'.+(?= (\[\')?\()', x).group()
            else:
                print(x)
                miles, yards = 0, 0
                add_info = ''
            length = measurement.measures.Distance(mi=miles).m + measurement.measures.Distance(yd=yards).m
        return length, add_info

    # Collect the data of railway tunnel lengths for a given page number
    def collect_railway_tunnel_lengths(self, page_no, update=False, verbose=False):
        """
        :param page_no: [int] page number; valid values include 1, 2, and 3
        :param update: [bool] indicate whether to re-scrape the tunnel lengths data for the given page_no
        :param verbose: [bool]
        :return [dict] containing:
                    [DataFrame] tunnel lengths data of the given 'page'
                    [str] date of when the data was last updated
        """
        assert page_no in range(1, 4)
        page_headers = self.find_page_headers()
        filename = fuzzywuzzy.process.extractOne(str(page_no), page_headers)[0]
        pickle_filename = re.sub(r"[()]", "", re.sub(r"[ -]", "-", filename)).lower() + ".pickle"

        path_to_pickle = self.cd_tunnels(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            tunnel_lengths_codes = load_pickle(path_to_pickle)

        else:
            url = self.URL.replace('tunnels0', 'tunnels{}'.format(page_no))

            try:
                last_updated_date = get_last_updated_date(url)
            except Exception as e:
                print("Failed to find the last updated date for tunnel lengths data on \"{}\". {}".format(filename, e))
                last_updated_date = ''

            try:
                source = requests.get(url)
                parsed_text = bs4.BeautifulSoup(source.text, 'lxml')  # Optional parsers:, 'html5lib', 'html.parser'

                header = [x.text for x in parsed_text.find_all('th')]  # Column names

                crossed = [re.match(r'^Between.*', x) for x in header]
                if any(crossed):
                    idx = list(itertools.compress(range(len(crossed)), crossed))
                    assert len(idx) == 1
                    header.remove(header[idx[0]])
                    header[idx[0]:idx[0]] = ['Station_O', 'Station_D']

                # Table data
                temp_tables = parsed_text.find_all('table')
                tbl_lst = parse_tr(header, trs=temp_tables[1].find_all('tr'))
                tbl_lst = [[item.replace('\r', ' ').replace('\xa0', '') for item in record] for record in tbl_lst]

                # Create a DataFrame
                tunnel_lengths_table = pd.DataFrame(data=tbl_lst, columns=header)
                tunnel_lengths_table[['Length_m', 'Length_note']] = \
                    tunnel_lengths_table.Length.map(self.parse_tunnel_length).apply(pd.Series)

            except Exception as e:
                print("Failed to collect tunnel lengths data on \"{}\". {}".format(filename, e))
                tunnel_lengths_table = pd.DataFrame()

            code_key = re.search(r'(?<=\()\w.*(?=\))', filename).group(0).replace('-', '_')
            tunnel_lengths_codes = {code_key: tunnel_lengths_table, 'Last_updated_date': last_updated_date}

            save_pickle(tunnel_lengths_codes, path_to_pickle, verbose)

        return tunnel_lengths_codes

    # Collect the data of minor lines and other odds / ends
    def collect_page4_others(self, update=False, verbose=False):
        """
        Page 4 (others) contains more than one table on the web page
        """
        page_headers = self.find_page_headers()
        filename = fuzzywuzzy.process.extractOne('others', page_headers)[0]
        pickle_filename = re.sub(r"[()]", "", re.sub(r"[ -]", "-", filename)).lower() + ".pickle"
        path_to_pickle = self.cd_tunnels(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            tunnel_lengths_data = load_pickle(path_to_pickle)

        else:
            url = self.HomeURL + '/tunnels/tunnels4.shtm'

            try:
                last_updated_date = get_last_updated_date(url)
            except Exception as e:
                print("Failed to find the last updated date for tunnel lengths data on \"{}\". {}".format(filename, e))
                last_updated_date = ''

            try:
                source = requests.get(url)
                parsed_text = bs4.BeautifulSoup(source.text, 'lxml')  # Optional parsers:, 'html5lib', 'html.parser'
                headers = []
                temp_header = parsed_text.find('table')
                while temp_header.find_next('th'):
                    header = [x.text for x in temp_header.find_all('th')]
                    if len(header) > 0:
                        crossed = [re.match('^Between.*', x) for x in header]
                        if any(crossed):
                            idx = list(itertools.compress(range(len(crossed)), crossed))
                            assert len(idx) == 1
                            header.remove(header[idx[0]])
                            header[idx[0]:idx[0]] = ['Station_O', 'Station_D']
                        headers.append(header)
                    temp_header = temp_header.find_next('table')

                tbl_lst = parsed_text.find_all('table')
                tbl_lst = operator.itemgetter(1, 3)(tbl_lst)
                tbl_lst = [parse_tr(header, x.find_all('tr')) for header, x in zip(headers, tbl_lst)]
                tbl_lst = [[[item.replace('\xa0', '') for item in record] for record in tbl] for tbl in tbl_lst]

                tunnel_lengths = [pd.DataFrame(tbl, columns=header) for tbl, header in zip(tbl_lst, headers)]
                for i in range(len(tunnel_lengths)):
                    tunnel_lengths[i][['Length_m', 'Length_note']] = \
                        tunnel_lengths[i].Length.map(self.parse_tunnel_length).apply(pd.Series)

                other_types = [x.text for x in parsed_text.find_all('h3')]

            except Exception as e:
                print("Failed to collect tunnel lengths data on \"{}\". {}".format(filename, e))
                other_types, tunnel_lengths = ['None'], pd.DataFrame()

            tunnel_lengths_data = dict(zip(other_types + ['Last_updated_date'], tunnel_lengths + [last_updated_date]))

            save_pickle(tunnel_lengths_data, path_to_pickle, verbose)

        return tunnel_lengths_data

    # Fetch all the collected data of railway tunnel lengths
    def fetch_railway_tunnel_lengths(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        """
        :param update: [bool]
        :param pickle_it: [bool]
        :param data_dir: [str; None]
        :param verbose: [bool]
        :return [dict] containing:
                    [DataFrame] railway tunnel lengths data, including the name,
                                length, owner and relative location
                    [str] date of when the data was last updated
        """
        data_sets = [self.collect_railway_tunnel_lengths(
            page_no, update, verbose=False if data_dir or not verbose else True) for page_no in range(1, 4)]
        other_data_set = self.collect_page4_others(update, verbose=False if data_dir or not verbose else True)

        tunnel_lengths_tables = [d[k] for d in data_sets for k, v in d.items() if k != 'Last_updated_date']
        others_tables = [v for k, v in other_data_set.items() if k != 'Last_updated_date']
        tunnel_lengths_tables += others_tables
        tunnel_lengths_codes = pd.concat(tunnel_lengths_tables, axis=0, ignore_index=True, sort=False)

        last_updated_dates = [d[k] for d in data_sets for k, v in d.items() if k == 'Last_updated_date']
        latest_update_date = max(last_updated_dates)

        tunnel_lengths = {'Tunnels': tunnel_lengths_codes, 'Latest_update_date': latest_update_date}

        if pickle_it and data_dir:
            self.CurrentDataDir = regulate_input_data_dir(data_dir)
            path_to_pickle = os.path.join(self.CurrentDataDir, "tunnel_lengths.pickle")
            save_pickle(tunnel_lengths, path_to_pickle, verbose=True)

        return tunnel_lengths



""" Depots """

import os
import re

import bs4
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.ops import confirmed
from pyhelpers.store import load_pickle

from pyrcs.utils import cd_dat, get_catalogue, get_last_updated_date
from pyrcs.utils import save_pickle


class Depots:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Depot codes'
        self.URL = self.HomeURL + '/depots/depots0.shtm'
        self.Catalogue = get_catalogue(self.URL)
        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("other-assets", "depots")
        self.CurrentDataDir = self.DataDir

    # Change directory to "dat\\other-assets\\depots\\"
    def cd_depots(self, *directories):
        path = self.DataDir
        for x in directories:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\other-assets\\depots\\dat"
    def cdd_sigbox(self, *sub_dir):
        path = self.cd_depots("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Collect 'Two character TOPS codes'
    def collect_two_char_tops_codes(self, confirmation_required=True, verbose=False):
        if confirmed("To collect two character TOPS codes?", confirmation_required=confirmation_required):
            title_name = list(self.Catalogue.keys())[1]
            url = self.Catalogue[title_name]

            try:
                header, two_char_tops_codes = pd.read_html(url, na_values=[''], keep_default_na=False)
                two_char_tops_codes.columns = header.columns.to_list()
                two_char_tops_codes.fillna('', inplace=True)

                last_updated_date = get_last_updated_date(url)

                two_char_tops_codes_data = {title_name.replace(' ', '_'): two_char_tops_codes,
                                            'Last_updated_date': last_updated_date}

                path_to_pickle = self.cd_depots(title_name.replace(" ", "-").lower() + ".pickle")
                save_pickle(two_char_tops_codes_data, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect \"{}\". {}".format(title_name, e))
                two_char_tops_codes_data = None

            return two_char_tops_codes_data

    # Fetch 'Two character TOPS codes'
    def fetch_two_char_tops_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        title_name = list(self.Catalogue.keys())[1]
        path_to_pickle = self.cd_depots(title_name.replace(" ", "-").lower() + ".pickle")

        if os.path.isfile(path_to_pickle) and not update:
            two_char_tops_codes_data = load_pickle(path_to_pickle)

        else:
            two_char_tops_codes_data = self.collect_two_char_tops_codes(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)
            if two_char_tops_codes_data:
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, os.path.basename(path_to_pickle))
                    save_pickle(two_char_tops_codes_data, path_to_pickle, verbose=True)
            else:
                print("No data of \"{}\" has been collected.".format(title_name))
        return two_char_tops_codes_data

    # Collect 'Four digit pre-TOPS codes'
    def collect_four_digit_pre_tops_codes(self, confirmation_required=True, verbose=False):
        if confirmed("To collect four digit pre-TOPS codes?", confirmation_required=confirmation_required):
            title_name = list(self.Catalogue.keys())[2]
            path_to_pickle = self.cd_depots(re.sub(r'[ -]', '-', title_name).lower() + ".pickle")

            url = self.Catalogue[title_name]

            try:
                source = requests.get(url)
                p_tags = bs4.BeautifulSoup(source.text, 'lxml').find_all('p')
                region_names = [x.text.replace('Jump to: ', '').strip().split(' | ')
                                for x in p_tags if x.text.startswith('Jump to: ')][0]

                data_sets = iter(pd.read_html(source.text, na_values=[''], keep_default_na=False))

                four_digit_pre_tops_codes_list = []
                for x in data_sets:
                    header, four_digit_pre_tops_codes = x, next(data_sets)
                    four_digit_pre_tops_codes.columns = header.columns.to_list()
                    four_digit_pre_tops_codes_list.append(four_digit_pre_tops_codes)

                last_updated_date = get_last_updated_date(url)

                four_digit_pre_tops_codes_data = {
                    re.sub(r'[ \-]', '_', title_name): dict(zip(region_names, four_digit_pre_tops_codes_list)),
                    'Last_updated_date': last_updated_date}
                save_pickle(four_digit_pre_tops_codes_data, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect \"{}\". {}".format(title_name, e))
                four_digit_pre_tops_codes_data = None

            return four_digit_pre_tops_codes_data

    # Fetch 'Four digit pre-TOPS codes'
    def fetch_four_digit_pre_tops_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        title_name = list(self.Catalogue.keys())[2]
        path_to_pickle = self.cd_depots(re.sub(r'[ -]', '-', title_name).lower() + ".pickle")

        if os.path.isfile(path_to_pickle) and not update:
            four_digit_pre_tops_codes_data = load_pickle(path_to_pickle)

        else:
            four_digit_pre_tops_codes_data = self.collect_four_digit_pre_tops_codes(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)
            if four_digit_pre_tops_codes_data:
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, os.path.basename(path_to_pickle))
                    save_pickle(four_digit_pre_tops_codes_data, path_to_pickle, verbose=True)
            else:
                print("No data of \"{}\" has been collected.".format(title_name))

        return four_digit_pre_tops_codes_data

    # Collect '1950 system (pre-TOPS) codes'
    def collect_1950_system_codes(self, confirmation_required=True, verbose=False):
        if confirmed("To collect 1950 system (pre-TOPS) codes?", confirmation_required=confirmation_required):
            title_name = list(self.Catalogue.keys())[3]
            url = self.Catalogue[title_name]

            try:
                header, system_1950_codes = pd.read_html(url, na_values=[''], keep_default_na=False)
                system_1950_codes.columns = header.columns.to_list()

                title_name_ = re.sub(r'[ -]', '_', re.sub(r'[()]', '', title_name))
                path_to_pickle = self.cd_depots(title_name_.replace("_", "-").lower() + ".pickle")

                last_updated_date = get_last_updated_date(url)

                system_1950_codes_data = {title_name_: system_1950_codes, 'Last_updated_date': last_updated_date}

                save_pickle(system_1950_codes_data, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect the data of \"{}\". {}".format(title_name, e))
                system_1950_codes_data = None

            return system_1950_codes_data

    # Fetch '1950 system (pre-TOPS) codes'
    def fetch_1950_system_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        title_name = list(self.Catalogue.keys())[3]
        title_name_ = re.sub(r'[ -]', '_', re.sub(r'[()]', '', title_name))
        path_to_pickle = self.cd_depots(title_name_.lower() + ".pickle")

        if os.path.isfile(path_to_pickle) and not update:
            system_1950_codes_data = load_pickle(path_to_pickle)

        else:
            system_1950_codes_data = self.collect_1950_system_codes(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)
            if system_1950_codes_data:
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, os.path.basename(path_to_pickle))
                    save_pickle(system_1950_codes_data, path_to_pickle, verbose=True)
            else:
                print("No data of \"{}\" has been collected.".format(title_name))

        return system_1950_codes_data

    # Collect 'GWR codes' - Great Western Railway depot codes
    def collect_gwr_codes(self, confirmation_required=True, verbose=False):
        if confirmed("To collect Great Western Railway depot codes?", confirmation_required=confirmation_required):
            title_name = list(self.Catalogue.keys())[4]
            url = self.Catalogue[title_name]

            try:
                header, gwr_codes = pd.read_html(url)
                gwr_codes.columns = header.columns.to_list()

                last_updated_date = get_last_updated_date(url)
                gwr_codes_data = {title_name.replace(' ', '_'): gwr_codes, 'Last_updated_date': last_updated_date}

                path_to_pickle = self.cd_depots(title_name.replace(" ", "-").lower() + ".pickle")
                save_pickle(gwr_codes_data, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect the data of \"{}.\" {}".format(title_name, e))
                gwr_codes_data = None

            return gwr_codes_data

    #
    def fetch_gwr_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        title_name = list(self.Catalogue.keys())[4]
        path_to_pickle = self.cd_depots(title_name.replace(" ", "-").lower() + ".pickle")

        if os.path.isfile(path_to_pickle) and not update:
            gwr_codes_data = load_pickle(path_to_pickle)

        else:
            gwr_codes_data = self.collect_gwr_codes(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)
            if gwr_codes_data:
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, os.path.basename(path_to_pickle))
                    save_pickle(gwr_codes_data, path_to_pickle, verbose=True)
            else:
                print("No data of \"{}\" has been collected.".format(title_name))

        return gwr_codes_data

    # Fetch all the collected depots codes
    def fetch_depot_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        two_char_tops_codes_data = self.fetch_two_char_tops_codes(update, verbose=verbose)
        four_digit_pre_tops_codes_data = self.fetch_four_digit_pre_tops_codes(update, verbose=verbose)
        system_1950_codes_data = self.fetch_1950_system_codes(update, verbose=verbose)
        gwr_codes_data = self.fetch_gwr_codes(update, verbose=verbose)

        depot_codes = {}
        for x in (two_char_tops_codes_data, four_digit_pre_tops_codes_data, system_1950_codes_data, gwr_codes_data):
            old_key = list(x.keys())[0]
            x['Data'] = x.pop(old_key)
            depot_codes.update({old_key: x})

        if pickle_it and data_dir:
            self.CurrentDataDir = regulate_input_data_dir(data_dir)
            path_to_pickle = os.path.join(self.CurrentDataDir, "depot-codes.pickle")
            save_pickle(depot_codes, path_to_pickle)

        return depot_codes



""" Features """

import copy
import os
import re
import unicodedata
import urllib.parse

import bs4
import numpy as np
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.ops import confirmed
from pyhelpers.store import load_pickle

from pyrcs.utils import cd_dat, get_last_updated_date
from pyrcs.utils import save_pickle


# Decode vulgar fraction
def decode_vulgar_fraction(string):
    for s in string:
        try:
            name = unicodedata.name(s)
            if name.startswith('VULGAR FRACTION'):
                # normalized = unicodedata.normalize('NFKC', s)
                # numerator, _, denominator = normalized.partition('⁄')
                # frac_val = int(numerator) / int(denominator)
                frac_val = unicodedata.numeric(s)
                return frac_val
        except (TypeError, ValueError):
            pass


# Parse 'VULGAR FRACTION'
def parse_vulgar_fraction_in_length(x):
    if x == '':
        yd = np.nan
    elif re.match(r'\d+yd', x):  # e.g. '620yd'
        yd = int(re.search(r'\d+(?=yd)', x).group(0))
    elif re.match(r'\d+&frac\d+;yd', x):  # e.g. '506&frac23;yd'
        yd, frac = re.search(r'([0-9]+)&frac([0-9]+)(?=;yd)', x).groups()
        yd = int(yd) + int(frac[0]) / int(frac[1])
    else:  # e.g. '557½yd'
        yd = decode_vulgar_fraction(x)
    return yd


class Features:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Infrastructure features'
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("other-assets", "features")
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "dat\\other-assets\\features\\"
    def cd_features(self, *directories):
        path = self.DataDir
        for x in directories:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\other-assets\\features\\dat"
    def cdd_features(self, *sub_dir):
        path = self.cd_features("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Collect 'hot axle box detectors (HABDs)' and 'wheel impact load detectors (WILDs)'
    def collect_habds_and_wilds(self, confirmation_required=True, verbose=False):

        if confirmed("To collect codes for HABDs and WILDs?", confirmation_required=confirmation_required):

            url = self.HomeURL + '/misc/habdwild.shtm'
            source = requests.get(url)
            titles = [x.text for x in bs4.BeautifulSoup(source.text, 'lxml').find_all('h3')]

            try:
                titles = [x[x.index('(') + 1:x.index(')')] for x in titles]
            except ValueError:
                pass

            try:
                habds_and_wilds_codes = iter(pd.read_html(url, na_values=[''], keep_default_na=False))

                habds_and_wilds_codes_list = []
                for x in habds_and_wilds_codes:
                    header, data = x, next(habds_and_wilds_codes)
                    data.columns = header.columns.to_list()
                    data.fillna('', inplace=True)
                    habds_and_wilds_codes_list.append(data)

                habds_and_wilds_codes_data = dict(zip(titles, habds_and_wilds_codes_list))

                last_updated_date = get_last_updated_date(url)
                habds_and_wilds_codes_data.update({'Last_updated_date': last_updated_date})

                path_to_pickle = self.cd_features("habds-and-wilds.pickle")
                save_pickle(habds_and_wilds_codes_data, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect the data of \"{}\". {}".format(' and '.join(titles), e))
                habds_and_wilds_codes_data = None

            return habds_and_wilds_codes_data

    # Fetch HABDs and WILDs
    def fetch_habds_and_wilds(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        pickle_filename = "habds-and-wilds.pickle"
        path_to_pickle = self.cd_features(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            habds_and_wilds_codes_data = load_pickle(path_to_pickle)

        else:
            habds_and_wilds_codes_data = self.collect_habds_and_wilds(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)

            if habds_and_wilds_codes_data:
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
                    save_pickle(habds_and_wilds_codes_data, path_to_pickle, verbose=True)
            else:
                print("No data of \"HABDs or WILDs\" has been collected.")

        return habds_and_wilds_codes_data

    # Collect 'Water troughs'
    def collect_water_troughs(self, confirmation_required=True, verbose=False):
        title_name = 'Water trough'

        if confirmed("To collect codes for \"{}\"?".format(title_name.lower()),
                     confirmation_required=confirmation_required):

            url = urllib.parse.urljoin(self.HomeURL, '/misc/troughs.shtm')

            try:
                header, water_troughs = pd.read_html(url)
                water_troughs.columns = header.columns.to_list()
                water_troughs.fillna('', inplace=True)
                water_troughs.Length = water_troughs.Length.map(parse_vulgar_fraction_in_length)
                water_troughs.rename(columns={'Length': 'Length_yard'}, inplace=True)

                last_updated_date = get_last_updated_date(url)
                water_troughs_data = {title_name.replace(' ', '_'): water_troughs,
                                      'Last_updated_date': last_updated_date}

                path_to_pickle = self.cd_features(title_name.lower().replace(" ", "-") + ".pickle")
                save_pickle(water_troughs_data, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect the data of \"{}\". {}".format(title_name.lower(), e))
                water_troughs_data = None

            return water_troughs_data

    # Fetch 'Water troughs'
    def fetch_water_troughs(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        title_name = 'Water trough'
        path_to_pickle = self.cd_features(title_name.lower().replace(" ", "-") + ".pickle")

        if os.path.isfile(path_to_pickle) and not update:
            water_troughs_data = load_pickle(path_to_pickle)

        else:
            water_troughs_data = self.collect_water_troughs(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)

            if water_troughs_data:
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, os.path.basename(path_to_pickle))
                    save_pickle(water_troughs_data, path_to_pickle, verbose=True)
            else:
                print("No data of \"{}\" has been collected.".format(title_name.lower()))

        return water_troughs_data

    # Collect 'Telegraph code words'
    def collect_telegraph_codes(self, confirmation_required=True, verbose=False):
        title_name = 'Telegraph code words'

        if confirmed("To collect codes for \"{}\"?".format(title_name.lower()),
                     confirmation_required=confirmation_required):

            url = urllib.parse.urljoin(self.HomeURL, '/misc/telegraph.shtm')

            try:
                source = requests.get(url)
                #
                sub_titles = [x.text for x in bs4.BeautifulSoup(source.text, 'lxml').find_all('h3')]
                #
                data_sets = iter(pd.read_html(source.text))
                telegraph_codes_list = []
                for x in data_sets:
                    header, telegraph_codes = x, next(data_sets)
                    telegraph_codes.columns = header.columns.to_list()
                    telegraph_codes_list.append(telegraph_codes)

                telegraph_codes_data = dict(zip(sub_titles, telegraph_codes_list))

                last_updated_date = get_last_updated_date(url)
                telegraph_codes_data.update({'Last_updated_date': last_updated_date})

                path_to_pickle = self.cd_features(title_name.lower().replace(" ", "-") + ".pickle")
                save_pickle(telegraph_codes_data, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect the data of \"{}.\" {}".format(title_name.lower(), e))
                telegraph_codes_data = None

            return telegraph_codes_data

    # Fetch 'Telegraph code words'
    def fetch_telegraph_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        title_name = 'Telegraph code words'
        path_to_pickle = self.cd_features(title_name.lower().replace(" ", "-") + ".pickle")

        if os.path.isfile(path_to_pickle) and not update:
            telegraph_codes_data = load_pickle(path_to_pickle)

        else:
            telegraph_codes_data = self.collect_telegraph_codes(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)

            if telegraph_codes_data:
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, os.path.basename(path_to_pickle))
                    save_pickle(telegraph_codes_data, path_to_pickle, verbose=True)
            else:
                print("No data of \"{}\" has been collected.".format(title_name.lower()))

        return telegraph_codes_data

    # Collect 'Buzzer codes'
    def collect_buzzer_codes(self, confirmation_required=True, verbose=False):
        title_name = 'Buzzer codes'

        if confirmed("To collect codes for \"{}\"?".format(title_name.lower()),
                     confirmation_required=confirmation_required):

            url = urllib.parse.urljoin(self.HomeURL, '/misc/buzzer.shtm')

            try:
                header, buzzer_codes = pd.read_html(url)
                buzzer_codes.columns = header.columns.to_list()
                buzzer_codes.fillna('', inplace=True)

                last_updated_date = get_last_updated_date(url)

                buzzer_codes_data = {'Codes': buzzer_codes, 'Last_updated_data': last_updated_date}

                path_to_pickle = self.cd_features(title_name.lower().replace(" ", "-") + ".pickle")
                save_pickle(buzzer_codes_data, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect the data of \"{}\". {}".format(title_name.lower(), e))
                buzzer_codes_data = None

            return buzzer_codes_data

    # Fetch 'Buzzer codes'
    def fetch_buzzer_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        title_name = 'Buzzer codes'
        path_to_pickle = self.cd_features(title_name.lower().replace(" ", "-") + ".pickle")

        if os.path.isfile(path_to_pickle) and not update:
            buzzer_codes_data = load_pickle(path_to_pickle)

        else:
            buzzer_codes_data = self.collect_buzzer_codes(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)

            if buzzer_codes_data:
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, os.path.basename(path_to_pickle))
                    save_pickle(buzzer_codes_data, path_to_pickle, verbose=True)
            else:
                print("No data of \"{}\" has been collected.".format(title_name.lower()))

        return buzzer_codes_data



"""

Data source: http://www.railwaycodes.org.uk

Railway track diagrams (http://www.railwaycodes.org.uk/track/diagrams0.shtm)

"""

import copy
import os
import urllib.parse

import bs4
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir

from pyrcs.utils import cd_dat, get_last_updated_date


class TrackDiagrams:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Railway track diagrams'
        self.URL = self.HomeURL + '/track/diagrams0.shtm'

        # Get contents
        source = requests.get(self.URL)
        soup, items = bs4.BeautifulSoup(source.text, 'lxml'), {}
        h3 = soup.find('h3', text=True, attrs={'class': None})
        while h3:
            # Description
            if h3.text == 'Miscellaneous':
                desc = [x.text for x in h3.find_next_siblings('p')]
            else:
                desc = h3.find_next_sibling('p').text.replace('\xa0', '')
            # Extract details
            cold_soup = h3.find_next('div', attrs={'class': 'columns'})
            if cold_soup:
                info = [x.text for x in cold_soup.find_all('p') if x.string != '\xa0']
                urls = [urllib.parse.urljoin(os.path.dirname(self.URL), x['href']) for x in cold_soup.find_all('a')]
            else:
                cold_soup = h3.find_next('a', attrs={'target': '_blank'})
                info, urls = [], []
                while cold_soup:
                    info.append(cold_soup.text)
                    urls.append(urllib.parse.urljoin(os.path.dirname(self.URL), cold_soup['href']))
                    cold_soup = cold_soup.find_next('a') if h3.text == 'Miscellaneous' \
                        else cold_soup.find_next_sibling('a')
            meta = pd.DataFrame(zip(info, urls), columns=['Description', 'FileURL'])
            items.update({h3.text: (desc, meta)})  # Update
            h3 = h3.find_next_sibling('h3')  # Move on
        self.Catalogue = items
        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("line-data", "track_diagrams")
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "...dat\\line-data\\track_diagrams\\" and sub-directories
    def cd_td(self, *sub_dir):
        path = self.DataDir
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\line-data\\track_diagrams\\dat" and sub-directories
    def cdd_td(self, *sub_dir):
        path = self.cd_td("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path



"""

Data source: http://www.railwaycodes.org.uk

Possession Resource Information Database (PRIDE)/Line Of Route (LOR) codes
(http://www.railwaycodes.org.uk/pride/pride0.shtm)

PRIDE numbers are now known as line of route (LOR) numbers.  The name has changed, but the codes are unchanged.

"""


import copy
import os
import re
import urllib.parse

import bs4
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.ops import confirmed
from pyhelpers.store import load_pickle

from pyrcs.utils import cd_dat, get_catalogue, get_last_updated_date, parse_tr
from pyrcs.utils import save_pickle


class LOR:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Line of Route (LOR/PRIDE) codes'
        self.URL = self.HomeURL + '/pride/pride0.shtm'
        self.Catalogue = get_catalogue(self.URL)
        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("line-data", "lor-codes")
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "dat\\line-data\\lor-codes" and sub-directories
    def cd_lor(self, *sub_dir):
        path = self.DataDir
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\line-data\\lor-codes\\dat" and sub-directories
    def cdd_lor(self, *sub_dir):
        path = self.cd_lor("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Get key to LOR code prefixes
    def get_key_to_prefixes(self, prefixes_only=True, update=False):
        path_to_pickle = self.cdd_lor("{}prefixes.pickle".format("" if prefixes_only else "key_to_"))
        if os.path.isfile(path_to_pickle) and not update:
            key_to_prefixes = load_pickle(path_to_pickle)
        else:
            try:
                lor_pref = pd.read_html(self.URL)[0].loc[:, [0, 2]]
                lor_pref.columns = ['Prefixes', 'Name']
                if prefixes_only:
                    key_to_prefixes = lor_pref.Prefixes.tolist()
                else:
                    key_to_prefixes = {'Key_to_prefixes': lor_pref, 'Last_update_date': self.Date}
                save_pickle(key_to_prefixes, path_to_pickle)
            except Exception as e:
                print("Failed to get the \"key to prefixes\". {}.".format(e))
                key_to_prefixes = [] if prefixes_only else {'Key_to_prefixes': None, 'Last_update_date': None}
        return key_to_prefixes

    # Get the urls to LOR codes with different prefixes
    def get_lor_page_urls(self, update=False):
        path_to_pickle = self.cdd_lor("urls.pickle")
        if os.path.isfile(path_to_pickle) and not update:
            urls = load_pickle(path_to_pickle)
        else:
            try:
                source = requests.get(self.URL)
                soup = bs4.BeautifulSoup(source.text, 'lxml')
                links = soup.find_all('a', href=re.compile('^pride|elrmapping'),
                                      text=re.compile('.*(codes|converter|Historical)'))
                urls = list(dict.fromkeys([self.URL.replace(os.path.basename(self.URL), x['href']) for x in links]))
                save_pickle(urls, path_to_pickle)
            except Exception as e:
                print("Failed to get the \"urls\" to LOR codes web pages. {}.".format(e))
                urls = []
        return urls

    # Update catalogue data
    def update_catalogue(self):
        if confirmed("To update catalogue?"):
            self.get_key_to_prefixes(prefixes_only=True, update=True)
            self.get_key_to_prefixes(prefixes_only=False, update=True)
            self.get_lor_page_urls(update=True)

    # Collect LOR codes by prefix
    def collect_lor_codes_by_prefix(self, prefixes, update=False, verbose=False):

        assert prefixes in self.get_key_to_prefixes(prefixes_only=True), \
            "\"prefixes\" must be one of {}".format(self.get_key_to_prefixes(prefixes_only=True))

        pickle_filename = "{}.pickle".format(prefixes if prefixes not in ("NW", "NZ") else "NW-NZ").lower()
        path_to_pickle = self.cd_lor("prefixes", pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            lor_codes_by_initials = load_pickle(path_to_pickle)

        else:
            try:
                prefixes = "NW" if prefixes in ("NW", "NZ") else prefixes
                url = self.HomeURL + '/pride/pride{}.shtm'.format(prefixes.lower())
                source = requests.get(url)
                source_text = source.text
                source.close()

                soup = bs4.BeautifulSoup(source_text, 'lxml')

                # Parse the column of Line Name
                def parse_line_name(x):
                    # re.search('\w+.*(?= \(\[\')', x).group(), re.search('(?<=\(\[\')\w+.*(?=\')', x).group()
                    try:
                        line_name, line_name_note = x.split(' ([\'')
                        line_name_note = line_name_note.strip('\'])')
                    except ValueError:
                        line_name, line_name_note = x, None
                    return line_name, line_name_note

                def parse_h3_table(tbl_soup):
                    header, code = tbl_soup
                    header_text = [h.text.replace('\n', ' ') for h in header.find_all('th')]
                    code_dat = pd.DataFrame(parse_tr(header_text, code.find_all('tr')), columns=header_text)
                    line_name_info = code_dat['Line Name'].map(parse_line_name).apply(pd.Series)
                    line_name_info.columns = ['Line Name', 'Line Name Note']
                    code_dat = pd.concat([code_dat, line_name_info], axis=1, sort=False)
                    try:
                        note_dat = dict([(x['id'].title(), x.text) for x in soup.find('ol').findChildren('a')])
                    except AttributeError:
                        note_dat = dict([('Note', None)])
                    return code_dat, note_dat

                h3, table_soup = soup.find_all('h3'), soup.find_all('table')
                if len(h3) == 0:
                    code_data, code_data_notes = parse_h3_table(table_soup)
                    lor_codes_by_initials = {'Code': code_data, 'Note': code_data_notes}
                else:
                    code_data_and_notes = [dict(zip(['Code', 'Note'], parse_h3_table(x)))
                                           for x in zip(*[iter(table_soup)] * 2)]
                    lor_codes_by_initials = dict(zip([x.text for x in h3], code_data_and_notes))

                last_updated_date = get_last_updated_date(url)
                lor_codes_by_initials.update({'Last_updated_date': last_updated_date})

                save_pickle(lor_codes_by_initials, path_to_pickle, verbose)

            except Exception as e:
                print("Failed to collect LOR codes with prefix \"{}\". {}".format(prefixes.upper(), e))
                lor_codes_by_initials = None

        return lor_codes_by_initials

    # Fetch all LOR codes either locally or from online
    def fetch_lor_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        prefixes = self.get_key_to_prefixes(prefixes_only=True, update=update)
        lor_codes = [self.collect_lor_codes_by_prefix(p, update, verbose) for p in prefixes if p != 'NZ']

        prefixes[prefixes.index('NW')] = 'NW-NZ'
        prefixes.remove('NZ')

        lor_codes_data = dict(zip(prefixes, lor_codes))

        # Get the latest updated date
        last_updated_dates = (item['Last_updated_date'] for item, _ in zip(lor_codes, prefixes))
        latest_update_date = max(d for d in last_updated_dates if d is not None)

        lor_codes_data.update({'Latest_update_date': latest_update_date})

        if pickle_it and data_dir:
            pickle_filename = "lor-codes.pickle"
            self.CurrentDataDir = regulate_input_data_dir(data_dir)
            path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
            save_pickle(lor_codes_data, path_to_pickle, verbose)

        return lor_codes_data

    # Collect ELR/LOR converter
    def collect_elr_lor_converter(self, confirmation_required=True, verbose=False):
        if confirmed("To collect ELR/LOR converter?", confirmation_required=confirmation_required):
            url = self.Catalogue['ELR/LOR converter']
            try:
                headers, elr_lor_dat = pd.read_html(url)
                elr_lor_dat.columns = list(headers)
                #
                source = requests.get(url)
                soup = bs4.BeautifulSoup(source.text, 'lxml')
                tds = soup.find_all('td')
                links = [x.get('href') for x in [x.find('a', href=True) for x in tds] if x is not None]
                elr_links, lor_links = [x for x in links[::2]], [x for x in links[1::2]]
                #
                if len(elr_links) != len(elr_lor_dat):
                    duplicates = elr_lor_dat[elr_lor_dat.duplicated(['ELR', 'LOR code'], keep=False)]
                    for i in duplicates.index:
                        if not duplicates['ELR'].loc[i].lower() in elr_links[i]:
                            elr_links.insert(i, elr_links[i - 1])
                        if not lor_links[i].endswith(duplicates['LOR code'].loc[i].lower()):
                            lor_links.insert(i, lor_links[i - 1])
                #
                elr_lor_dat['ELR_URL'] = [urllib.parse.urljoin(self.HomeURL, x) for x in elr_links]
                elr_lor_dat['LOR_URL'] = [self.HomeURL + 'pride/' + x for x in lor_links]
                #
                elr_lor_converter = {'ELR_LOR_converter': elr_lor_dat, 'Last_updated_date': get_last_updated_date(url)}

                save_pickle(elr_lor_converter, self.cd_lor("elr-lor-converter.pickle"), verbose)

            except Exception as e:
                print("Failed to collect \"ELR/LOR converter\". {}".format(e))
                elr_lor_converter = None

            return elr_lor_converter

    # Fetch ELR/LOR converter
    def fetch_elr_lor_converter(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        pickle_filename = "elr-lor-converter.pickle"
        path_to_pickle = self.cd_lor(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            elr_lor_converter = load_pickle(path_to_pickle)

        else:
            elr_lor_converter = self.collect_elr_lor_converter(confirmation_required=False,
                                                               verbose=False if data_dir or not verbose else True)
            if elr_lor_converter:  # codes_for_ole is not None
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
                    save_pickle(elr_lor_converter, path_to_pickle, verbose=True)
            else:
                print("No data of \"ELR/LOR converter\" has been collected for national network OLE installations.")

        return elr_lor_converter



"""

Data source: http://www.railwaycodes.org.uk

Railway line names (http://www.railwaycodes.org.uk/misc/line_names.shtm)

"""

import copy
import os
import re

import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.ops import confirmed
from pyhelpers.store import load_pickle

from pyrcs.utils import cd_dat, get_last_updated_date, parse_table, save_pickle


class LineNames:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Railway line names'
        self.URL = self.HomeURL + '/misc/line_names.shtm'
        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("line-data", "line-names")
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "dat\\line-data\\line-names" and sub-directories
    def cd_ln(self, *sub_dir):
        path = self.DataDir
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\line-data\\line-names\\dat" and sub-directories
    def cdd_ln(self, *sub_dir):
        path = self.cd_ln("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Scrape the data of railway line names
    def collect_line_names(self, confirmation_required=True, verbose=False):
        """
        :return [tuple] {'Line_names': [DataFrame] railway line names and routes data,
                         'Last_updated_date': [str] date of when the data was last updated}
        """
        if confirmed("To collect line names?", confirmation_required=confirmation_required):

            try:
                last_updated_date = get_last_updated_date(self.URL)
            except Exception as e:
                print("Failed to find the last update date for line names. {}".format(e))
                last_updated_date = ''

            try:
                source = requests.get(self.URL)
                row_lst, header = parse_table(source, parser='lxml')
                line_names = pd.DataFrame([[r.replace('\xa0', '').strip() for r in row] for row in row_lst],
                                          columns=header)

                # Parse route column
                def parse_route_column(x):
                    if 'Watford - Euston suburban route' in x:
                        route, route_note = 'Watford - Euston suburban route', x
                    elif ', including Moorgate - Farringdon' in x:
                        route_note = 'including Moorgate - Farringdon'
                        route = x.replace(', including Moorgate - Farringdon', '')
                    elif re.match(r'.+(?= \[\')', x):
                        route, route_note = re.split(r' \[\'\(?', x)
                        route_note = route_note.strip(")']")
                    elif re.match(r'.+\)$', x):
                        if re.match(r'.+(?= - \()', x):
                            route, route_note = x, None
                        else:
                            route, route_note = re.split(r' \(\[?\'?', x)
                            route_note = route_note.rstrip('\'])')
                    else:
                        route, route_note = x, None
                    return route, route_note

                line_names[['Route', 'Route_note']] = line_names.Route.map(parse_route_column).apply(pd.Series)

                line_names_data = {'Line_names': line_names, 'Last_updated_date': last_updated_date}

                save_pickle(line_names_data, self.cd_ln("line-names.pickle"), verbose)

            except Exception as e:
                print("Failed to collect line names. {}".format(e))
                line_names_data = None

            return line_names_data

    # Get the data of line names either locally or from online
    def fetch_line_names(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        pickle_filename = "line-names.pickle"
        path_to_pickle = self.cd_ln(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            line_names_data = load_pickle(path_to_pickle)

        else:
            line_names_data = self.collect_line_names(confirmation_required=False,
                                                      verbose=False if data_dir or not verbose else True)
            if line_names_data:  # line-names is not None
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
                    save_pickle(line_names_data, path_to_pickle, verbose=True)
            else:
                print("No data of the railway line names has been collected.")

        return line_names_data



""" Signal box prefix codes """

import copy
import os
import string

import bs4
import pandas as pd
import requests
from pyhelpers.dir import regulate_input_data_dir
from pyhelpers.ops import confirmed
from pyhelpers.store import load_pickle

from pyrcs.utils import cd_dat, get_catalogue, get_last_updated_date, parse_table, parse_tr
from pyrcs.utils import save_pickle


class SignalBoxes:
    def __init__(self, data_dir=None):
        self.HomeURL = 'http://www.railwaycodes.org.uk'
        self.Name = 'Signal box prefix codes'
        self.URL = self.HomeURL + '/signal/signal_boxes0.shtm'
        self.Catalogue = get_catalogue(self.URL)
        self.Date = get_last_updated_date(self.URL, parsed=True, date_type=False)
        self.DataDir = regulate_input_data_dir(data_dir) if data_dir else cd_dat("other-assets", "signal-boxes")
        self.CurrentDataDir = copy.copy(self.DataDir)

    # Change directory to "dat\\other-assets\\signal-boxes\\"
    def cd_sigbox(self, *directories):
        path = self.DataDir
        for x in directories:
            path = os.path.join(path, x)
        return path

    # Change directory to "dat\\other-assets\\signal-boxes\\dat"
    def cdd_sigbox(self, *sub_dir):
        path = self.cd_sigbox("dat")
        for x in sub_dir:
            path = os.path.join(path, x)
        return path

    # Collect signal box prefix codes for the given 'keyword' (a starting letter)
    def collect_signal_box_prefix_codes(self, initial, update=False, verbose=False):
        """
        :param initial: [str]
        :param update: [bool]
        :param verbose: [bool]
        :return: [dict; None]
        """
        path_to_pickle = self.cd_sigbox("a-z", initial.lower() + ".pickle")

        if os.path.isfile(path_to_pickle) and not update:
            signal_box_prefix_codes = load_pickle(path_to_pickle)

        else:
            sig_keys = [initial.upper(), 'Last_updated_date']

            if initial.upper() not in list(self.Catalogue.keys()):
                print("No data is available for signal box codes beginning with \"{}\".".format(initial.upper()))
                signal_box_prefix_codes = dict(zip(sig_keys, [pd.DataFrame(), '']))

            else:
                url = self.URL.replace('0', initial.lower())

                try:
                    source = requests.get(url)
                    # Get table data and its column names
                    records, header = parse_table(source, parser='lxml')
                    header = [h.replace('Signal box', 'Signal Box') for h in header]
                    # Create a DataFrame of the requested table
                    signal_boxes_data_table = pd.DataFrame([[x.strip('\xa0') for x in i] for i in records],
                                                           columns=header)

                except IndexError:
                    print("Failed to collect signal box prefix codes beginning with \"{}\".".format(initial.upper()))
                    signal_boxes_data_table = pd.DataFrame()

                try:
                    last_updated_date = get_last_updated_date(url)
                except Exception as e:
                    print("Failed to find the last updated date of the signal boxes codes beginning with \"{}\". "
                          "{}".format(initial.upper(), e))
                    last_updated_date = ''

                signal_box_prefix_codes = dict(zip(sig_keys, [signal_boxes_data_table, last_updated_date]))

            save_pickle(signal_box_prefix_codes, path_to_pickle, verbose)

        return signal_box_prefix_codes

    # Fetch all of the collected signal box prefix codes
    def fetch_signal_box_prefix_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        """
        :param update: [bool]
        :param pickle_it: [bool]
        :param data_dir: [str; None]
        :param verbose: [bool]
        :return: [dict; None]
        """
        # Get every data table
        data = [self.collect_signal_box_prefix_codes(x, update, verbose=False if data_dir or not verbose else True)
                for x in string.ascii_lowercase]

        # Select DataFrames only
        signal_boxes_data = (item[x] for item, x in zip(data, string.ascii_uppercase))
        signal_boxes_data_table = pd.concat(signal_boxes_data, axis=0, ignore_index=True, sort=False)

        # Get the latest updated date
        last_updated_dates = (item['Last_updated_date'] for item in data)
        latest_update_date = max(d for d in last_updated_dates if d != '')

        # Create a dict to include all information
        signal_box_prefix_codes = {'Signal_boxes': signal_boxes_data_table, 'Latest_update_date': latest_update_date}

        if pickle_it and data_dir:
            self.CurrentDataDir = regulate_input_data_dir(data_dir)
            path_to_pickle = os.path.join(self.CurrentDataDir, "signal_box_prefix_codes.pickle")
            save_pickle(signal_box_prefix_codes, path_to_pickle, verbose=True)

        return signal_box_prefix_codes

    # Collect non-national rail codes
    def collect_non_national_rail_codes(self, confirmation_required=True, verbose=False):
        if confirmed("To collect non-national rail signal box prefix codes?",
                     confirmation_required=confirmation_required):
            url = self.HomeURL + '/signal/signal_boxesX.shtm'
            try:
                source = requests.get(url)
                web_page_text = bs4.BeautifulSoup(source.text, 'lxml')
                non_national_rail, non_national_rail_codes = [], {}

                for h in web_page_text.find_all('h3'):
                    non_national_rail_name = h.text  # Get the name of the non-national rail

                    # Find text descriptions
                    desc = h.find_next('p')
                    desc_text, more_desc = desc.text.replace('\xa0', ''), desc.find_next('p')
                    while more_desc.find_previous('h3') == h:
                        desc_text = '\n'.join([desc_text, more_desc.text.replace('\xa0', '')])
                        more_desc = more_desc.find_next('p')
                        if more_desc is None:
                            break

                    # Get table data
                    tbl_dat = desc.find_next('table')
                    if tbl_dat.find_previous('h3').text == non_national_rail_name:
                        header = [th.text for th in tbl_dat.find_all('th')]  # header
                        data = pd.DataFrame(parse_tr(header, tbl_dat.find_next('table').find_all('tr')), columns=header)
                    else:
                        data = None

                    # Update data dict
                    non_national_rail_codes.update(
                        {non_national_rail_name: {'Codes': data, 'Note': desc_text.replace('\xa0', '')}})

                save_pickle(non_national_rail_codes, self.cd_sigbox("non-national-rail-signals.pickle"), verbose)

            except Exception as e:
                print("Failed to collect non-national rail codes. {}".format(e))
                non_national_rail_codes = None

            return non_national_rail_codes

    # Fetch the collected non-national rail codes
    def fetch_non_national_rail_codes(self, update=False, pickle_it=False, data_dir=None, verbose=False):
        pickle_filename = "non-national-rail-signals.pickle"
        path_to_pickle = self.cd_sigbox(pickle_filename)

        if os.path.isfile(path_to_pickle) and not update:
            non_national_rail_codes = load_pickle(path_to_pickle)

        else:
            non_national_rail_codes = self.collect_non_national_rail_codes(
                confirmation_required=False, verbose=False if data_dir or not verbose else True)

            if non_national_rail_codes:
                if pickle_it and data_dir:
                    self.CurrentDataDir = regulate_input_data_dir(data_dir)
                    path_to_pickle = os.path.join(self.CurrentDataDir, pickle_filename)
                    save_pickle(non_national_rail_codes, path_to_pickle)
            else:
                print("No data of non-national rail codes has been collected.")

        return non_national_rail_codes
